#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Create a KarhunenLoeveResult
mesh = ot.IntervalMesher([9]).build(ot.Interval(-1.0, 1.0))
cov1D = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov1D, 0.0)
algo.run()
result = algo.getResult()
projection = ot.KarhunenLoeveProjection(result)
# Construction based on a FieldFunction followed by a FieldToPointFunction
fieldFunction = ot.ValueFunction(ot.SymbolicFunction("x", "x"), mesh)
# Create an instance
myFunc = ot.FieldToPointConnection(projection, fieldFunction)

print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a field
field = result.getModesAsProcessSample().computeMean()
print("field=", field)
print("myFunc(field)=", myFunc(field.getValues()))
print("called ", myFunc.getCallsNumber(), " times")
# Connection on a sample
sample = result.getModesAsProcessSample()
print("sample=", sample)
Y = myFunc(sample)
print("myFunc(sample)=", len(Y), Y.getDimension())
print("called ", myFunc.getCallsNumber(), " times")

# Construction based on a FieldToPointFunction followed by a Function
dimension = projection.getOutputDimension()
function = ot.SymbolicFunction(
    ot.Description.BuildDefault(dimension, "x"),
    ot.Description.BuildDefault(dimension, "x"),
)
# Create an instance
myFunc = ot.FieldToPointConnection(function, projection)

print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a field
field = result.getModesAsProcessSample().computeMean()
print("field=", field)
print("myFunc(field)=", myFunc(field.getValues()))
# Get the number of calls
print("called ", myFunc.getCallsNumber(), " times")

# 2-d mesh
n = 5
indices = [n, n]
intervalMesher = ot.IntervalMesher(indices)
interval = ot.Interval([0.0] * 2, [1.0] * 2)
mesh2D = intervalMesher.build(interval)


def f2Pfunc(X):
    Y = ot.Sample(X).computeMean()
    return Y


field2PFunction = ot.PythonFieldToPointFunction(mesh2D, 1, 1, f2Pfunc)
fieldFunction = ot.ValueFunction(ot.SymbolicFunction(["x", "y"], ["3*x"]), mesh2D)
myFunc = ot.FieldToPointConnection(field2PFunction, fieldFunction)
print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
field = ot.Field(mesh2D, ot.Normal(2).getSample(mesh2D.getVerticesNumber()))
print("myFunc(field)=", myFunc(field))
print("called ", myFunc.getCallsNumber(), " times")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

mesh = ot.IntervalMesher([9]).build(ot.Interval(-1.0, 1.0))
cov1D = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov1D, 0.0)
algo.run()
result = algo.getResult()
projection = ot.KarhunenLoeveProjection(result)
print("projection=", projection)
# Get the input and output description
print("projection input description=", projection.getInputDescription())
print("projection output description=", projection.getOutputDescription())
# Get the spatial, input and output dimension
print("projection input dimension=", projection.getInputMesh().getDimension())
print("projection input dimension=", projection.getInputDimension())
print("projection output dimension=", projection.getOutputDimension())
field = result.getModesAsProcessSample().computeMean()
point = projection(field)
print("field=", field)
print("point=", point)
# Get the number of calls
print("called ", projection.getCallsNumber(), " times")



import matplotlib.pyplot as plt

# Create the figure
plt.figure(1, figsize=(4, 4))
ax = plt.subplot(111)

# Create the points
ax.plot(
    [0, 0, 1, 0, 1, 2, 3, 2, 1, 0], [0, 1, 0, 2, 1, 0, 0, 1, 2, 3], "o", markersize=9
)

# Create the arrows
ax.annotate(
    "",
    xy=(0.97, 0),
    xycoords="data",
    xytext=(0, 0),
    textcoords="data",
    arrowprops=dict(
        arrowstyle="-|>",
        linestyle="dashed",
        mutation_scale=15,
        connectionstyle="arc3",
        color="black",
    ),
)

ax.arrow(
    1,
    0,
    -0.97,
    0.97,
    head_width=0.08,
    head_length=0.08,
    fc="k",
    ec="k",
    length_includes_head=True,
    linestyle="dashed",
)
ax.arrow(
    0,
    1,
    1.97,
    -0.97,
    head_width=0.08,
    head_length=0.08,
    fc="k",
    ec="k",
    length_includes_head=True,
    linestyle="dashed",
)
ax.arrow(
    2,
    0,
    -1.97,
    1.97,
    head_width=0.08,
    head_length=0.08,
    fc="k",
    ec="k",
    length_includes_head=True,
    linestyle="dashed",
)
ax.arrow(
    0,
    2,
    2.97,
    -1.97,
    head_width=0.08,
    head_length=0.08,
    fc="k",
    ec="k",
    length_includes_head=True,
    linestyle="dashed",
)
ax.arrow(
    3,
    0,
    -2.97,
    2.97,
    head_width=0.08,
    head_length=0.08,
    fc="k",
    ec="k",
    length_includes_head=True,
    linestyle="dashed",
)
ax.arrow(
    0,
    3,
    1.97,
    -0.97,
    head_width=0.08,
    head_length=0.08,
    fc="k",
    ec="k",
    length_includes_head=True,
    linestyle="dashed",
)

# Annotate points
ax.annotate(
    "4",
    xy=(1, 1),
    xycoords="data",
    xytext=(-20, -5),
    textcoords="offset points",
    fontsize=16,
)

ax.annotate(
    "7",
    xy=(2, 1),
    xycoords="data",
    xytext=(+15, +0),
    textcoords="offset points",
    fontsize=16,
)

ax.annotate(
    "8",
    xy=(1, 2),
    xycoords="data",
    xytext=(+15, +0),
    textcoords="offset points",
    fontsize=16,
)

# Add labels
ax.annotate(
    r"$\tau_1$",
    xy=(1, 0),
    xytext=(10, 10),
    ha="left",
    va="center",
    xycoords="axes fraction",
    textcoords="offset points",
    fontsize=20,
)

ax.annotate(
    r"$\tau_2$",
    xy=(0, 1),
    xytext=(0, 10),
    ha="left",
    va="center",
    xycoords="axes fraction",
    textcoords="offset points",
    fontsize=20,
)

# Hide spines
ax.spines["right"].set_color("none")
ax.spines["top"].set_color("none")

# Set spines's position
ax.xaxis.set_ticks_position("bottom")
ax.spines["bottom"].set_position(("data", 0))
ax.yaxis.set_ticks_position("left")
ax.spines["left"].set_position(("data", 0))

# Add labels
plt.xticks([-0.1] + list(range(4)) + [3.2])
ax.set_xticklabels(("", "$0$", "$1$", "$3$", "$6$", ""), fontsize=20)
plt.yticks([-0.1] + list(range(4)) + [3.2])
ax.set_yticklabels(("", "", "$2$", "$5$", "$9$", ""), fontsize=20)

# Show the figure
plt.show()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Dimension of the input model
# Size of the TimeGrid
size = 64
dimension = 1
timeGrid = ot.RegularGrid(0.0, 0.1, size)
amplitude = ot.Point(dimension, 2.0)
scale = ot.Point(dimension, 1.0)
model = ot.CauchyModel(scale, amplitude)
myProcess = ot.SpectralGaussianProcess(model, timeGrid)

# Create a Process sample
N = 100
sample = ot.ProcessSample(myProcess.getSample(N))

# Filtering Windows
myFactory = ot.WelchFactory()

# Build a UserDefinedSpectralModel using the Wellch method
mySpectralModel = myFactory.buildAsUserDefinedSpectralModel(sample)

# Get the frequency grid of the model
myFrequencyGrid = mySpectralModel.getFrequencyGrid()
for i in range(dimension):
    for j in range(dimension):
        print("Spectre ", i, "-", j)
        for k in range(myFrequencyGrid.getN()):
            frequency = myFrequencyGrid.getStart() + k * myFrequencyGrid.getStep()
            estimatedValue = (mySpectralModel(frequency)[i, j]).real
            modelValue = (model(frequency)[i, j]).real
            print(
                "Frequency =  %.6f" % frequency,
                ", evaluation = %.8f" % estimatedValue,
                " model = %.8f" % modelValue,
            )

# Create a Time Series
timeSeries = myProcess.getRealization()

# Build a UserDefinedSpectralModel using the Wellch method
mySpectralModel2 = myFactory.buildAsUserDefinedSpectralModel(timeSeries)

# Get the frequency grid of the model
myFrequencyGrid = mySpectralModel2.getFrequencyGrid()
for i in range(dimension):
    for j in range(dimension):
        print("Spectre ", i, "-", j)
        for k in range(myFrequencyGrid.getN()):
            frequency = myFrequencyGrid.getStart() + k * myFrequencyGrid.getStep()
            estimatedValue = (mySpectralModel2(frequency)[i, j]).real
            modelValue = (model(frequency)[i, j]).real
            print(
                "Frequency =  %.6f" % frequency,
                ", evaluation = %.8f" % estimatedValue,
                " model = %.8f" % modelValue,
            )



"""
Sampling from an unscaled probability density
=============================================
"""

# %%
# In this example, we show different ways to sample a distribution which PDF is known up to a normalization constant:
#
# - Case 1: computing the normalization factor and using a :class:`~openturns.PythonDistribution`,
# - Case 2: using the Ratio of Uniforms algorithm with the class :class:`~openturns.experimental.RatioOfUniforms`
# - Case 3: using some Metropolis-Hastings algorithms with the class :class:`~openturns.IndependentMetropolisHastings`
#   and :class:`~openturns.RandomWalkMetropolisHastings`.
#
# Consider a distribution whose
# probability density function :math:`p` is known up to the normalization factor :math:`c`:
# let :math:`f` be a function such that
# :math:`p = cf` with :math:`c \in \Rset^+_*`.
#
# We illustrate the case with:
#
# .. math::
#
#     f(x) = \frac{1}{2} (2 + \sin(x)^2) \exp \left[- \left(2 + \cos(3x)^3 + \sin(2x)^3 \right) x
#     \right]  \mathbf{1}_{[0, 2 \pi]}(x).
#
# First, we draw the unscaled probability density function.

import openturns as ot
import openturns.experimental as otexp
import openturns.viewer as otv
from math import pi
from time import monotonic

ot.RandomGenerator.SetSeed(1)
f = ot.SymbolicFunction(
    "x", "0.5 * (2 + sin(x)^2) * exp( -( 2 + cos(3*x)^3 + sin(2*x)^3) * x )"
)
lower_bound = 0.0
upper_bound = 2.0 * pi
range_PDF = ot.Interval(lower_bound, upper_bound)
graph = f.draw(lower_bound, upper_bound, 512)
graph.setTitle(
    r"Christian Robert function: $f(x) =  0.5(2 + sin^2 x) e^{ -x( 2 + cos^33x + sin^3 2x)}, \quad x \in [0, 2\pi]$"
)
graph.setXTitle("x")
graph.setYTitle(r"$f(x)$")
view = otv.View(graph)


# %%
# Case 1: Computation of the normalization factor
# -----------------------------------------------
# The best thing to do is to compute the normalization factor thanks to the integration algorithms of the library.
#
# We show how to compute the normalization factor using a :class:`~openturns.GaussKronrod` quadrature formula.
denom_fact = ot.GaussKronrod().integrate(f, range_PDF)[0]
norm_fact = 1.0 / denom_fact
print("normalization factor = ", norm_fact)

# %%
# Thus, we can define the exact PDF expression:
#
# .. math::
#
#   p(x) = \dfrac{f(x)}{\int_0^{2\pi} f(u)\, du}
#
exact_PDF = ot.LinearCombinationFunction([f], [norm_fact])

# %%
# Then we define a :class:`~openturns.PythonDistribution` which:
#
# - *computePDF* is computed from the exact expression,
# - *computeCDF* is computed using an integration algorithm on the *computePDF*.
#
# Doing that way, we use the generic sampler of the distribution, based on the CDF inversion method,
# implemented in the *getSample* method.


class NewDistribution(ot.PythonDistribution):
    def __init__(self):
        super().__init__(1)
        self.PDF_ = exact_PDF
        self.logPDF_ = ot.ComposedFunction(
            ot.SymbolicFunction("x", "log(x)"), self.PDF_
        )
        self.rangeLow_ = 0.0
        self.rangeUp_ = 2.0 * pi

    def getRange(self):
        return ot.Interval(self.rangeLow_, self.rangeUp_)

    def computeLogPDF(self, X):
        if X[0] < self.rangeLow_ or X[0] >= self.rangeUp_:
            return -ot.SpecFunc.Infinity
        return self.logPDF_(X)[0]

    def computePDF(self, X):
        if X[0] < self.rangeLow_ or X[0] >= self.rangeUp_:
            return 0.0
        return self.PDF_(X)[0]

    def computeCDF(self, X):
        if X[0] < self.rangeLow_:
            return 0.0
        if X[0] >= self.rangeUp_:
            return 1.0
        return ot.GaussLegendre([64]).integrate(
            self.PDF_, ot.Interval(self.rangeLow_, X[0])
        )[0]


# %%
# We create the new distribution that uses the generic sampler:
newDist_generic = ot.Distribution(NewDistribution())

# %%
# We can draw the exact PDF and the PDF estimated from a sample generated by the Ratio of Uniforms algorithm.

# %%
size = 500
sample = newDist_generic.getSample(size)

ks_algo = ot.KernelSmoothing()
ks_algo.setBoundaryCorrection(True)
ks_algo.setLowerBound(lower_bound)
ks_algo.setUpperBound(upper_bound)
ks_pdf_GS = ks_algo.build(sample)

g = ks_pdf_GS.drawPDF(-0.5, upper_bound * 1.1, 1001)
draw_exact = exact_PDF.draw(lower_bound, upper_bound, 1001).getDrawable(0)
draw_exact.setLineWidth(2)
g.add(draw_exact)
g.setLegends(["GS size = " + str(size), "exact pdf"])
g.setLegendPosition("upper right")
g.setTitle("Generic sampler (GS)")
g.setXTitle("x")
view = otv.View(g)

# %%
# We can change the sampler of that new distribution by implementing the method *getSample* and *getRealization* as follows:


class NewDistribution_RoU(ot.PythonDistribution):
    def __init__(self):
        super().__init__(1)
        self.PDF_ = exact_PDF
        self.logPDF_ = ot.ComposedFunction(
            ot.SymbolicFunction("x", "log(x)"), self.PDF_
        )
        self.rangeLow_ = 0.0
        self.rangeUp_ = 2.0 * pi
        self.sampler_ = otexp.RatioOfUniforms(self.logPDF_, self.getRange())

    def getRange(self):
        return ot.Interval(self.rangeLow_, self.rangeUp_)

    def computeLogPDF(self, X):
        if X[0] < self.rangeLow_ or X[0] >= self.rangeUp_:
            return -ot.SpecFunc.Infinity
        return self.logPDF_(X)[0]

    def computePDF(self, X):
        if X[0] < self.rangeLow_ or X[0] >= self.rangeUp_:
            return 0.0
        return self.PDF_(X)[0]

    def computeCDF(self, X):
        if X[0] < self.rangeLow_:
            return 0.0
        if X[0] >= self.rangeUp_:
            return 1.0
        return ot.GaussLegendre([32]).integrate(
            self.PDF_, ot.Interval(self.rangeLow_, X[0])
        )[0]

    def getRealization(self):
        return self.sampler_.getRealization()

    def getSample(self, n):
        return self.sampler_.getSample(n)


# %%
# We create the new distribution that uses the Ratio of Uniforms sampler:
newDist_RoU = ot.Distribution(NewDistribution_RoU())

# %%
# We compare the sampling speed of the distribution with the Ratio of Uniforms algorithm to the generic sampling speed.
# The Ratio of Uniforms algorithm proves to be much quicker than the generic method.
sizeRoU = 10000
sizeGeneric = 100

t0 = monotonic()
sample_RoU = newDist_RoU.getSample(sizeRoU)
t1 = monotonic()
sample_generic = newDist_generic.getSample(sizeGeneric)
t2 = monotonic()

speedRoU = sizeRoU / (t1 - t0)
speedGeneric = sizeGeneric / (t2 - t1)
print(f"Is Ratio of Uniforms faster ? {speedRoU > 10 * speedGeneric}")

# %%
# Case 2: Direct use the Ratio of Uniforms algorithm
# --------------------------------------------------
# In that case, we want to use the :class:`~openturns.experimental.RatioOfUniforms` algorithm to
# sample :math:`p`. We need to compute the function :math:`\log f` and its range.
#
# We create the function :math:`\log f` and the :class:`~openturns.experimental.RatioOfUniforms`:
log_UnscaledPDF = ot.ComposedFunction(ot.SymbolicFunction("x", "log(x)"), f)
ratio_algo = otexp.RatioOfUniforms(log_UnscaledPDF, range_PDF)

# %%
# We can draw the exact PDF and the PDF estimated from a sample generated by the Ratio of Uniforms algorithm.

# %%
size = 100000
sample = ratio_algo.getSample(size)

ks_algo = ot.KernelSmoothing()
ks_algo.setBoundaryCorrection(True)
ks_algo.setLowerBound(lower_bound)
ks_algo.setUpperBound(upper_bound)
ks_pdf = ks_algo.build(sample)

g = ks_pdf.drawPDF(-0.5, upper_bound * 1.1, 1001)
g.add(draw_exact)
g.setLegends(["RoU size = " + str(size), "exact PDF"])
g.setLegendPosition("upper right")
g.setTitle("Ratio of Uniforms (RoU) ")
g.setXTitle("x")
view = otv.View(g)

# %%
# In order to facilitate the comparison with the use of Metropolis-Hastings
# based algorithms, we generate
# a sample of the same size and we draw the estimated PDF.
size2 = 500
sample = ratio_algo.getSample(size2)

ks_algo = ot.KernelSmoothing()
ks_algo.setBoundaryCorrection(True)
ks_algo.setLowerBound(lower_bound)
ks_algo.setUpperBound(upper_bound)
ks_pdf_RoU = ks_algo.build(sample)

draw_RoU = ks_pdf_RoU.drawPDF(-0.5, upper_bound * 1.1, 1001).getDrawable(0)
draw_RoU.setLineWidth(2)
g.add(draw_RoU)
g.setLegends(
    ["RoU size = " + str(size), "exact PDF", "RoU (size = " + str(size2) + ")"]
)
g.setTitle("Ratio of Uniforms")
view = otv.View(g)

# %%
# By default, the parameter :math:`r=1`. We can get the associated acceptance ratio:
print("Acceptance ratio = ", ratio_algo.getAcceptanceRatio())

# %%
# We can estimate the normalization factor with the  Ratio of Uniforms algorithm (see
# the documenttaion of :class:`~openturns.experimental.RatioOfUniforms`):
print("Normalization factor estimate = ", ratio_algo.getC())

# %%
# We can change the :math:`r` parameter and check the associated acceptance ratio:
r_new = 0.5
ratio_algo.setR(r_new)
print("New acceptance ratio = ", ratio_algo.getAcceptanceRatio())

# %%
# Case 3(a): Use of Independent Metropolis-Hastings
# -------------------------------------------------
# Let us use a mixture distribution to approximate the target distribution.
#
# This approximation will serve as the instrumental distribution
# in the independent Metropolis-Hastings algorithm.

exp = ot.Exponential(1.0)
unif = ot.Normal(5.3, 0.4)
instrumental_dist = ot.Mixture([exp, unif], [0.9, 0.1])

# %%
# Compare the instrumental density to the exact PDF.
graph = instrumental_dist.drawPDF(lower_bound, upper_bound, 512)
graph.add(draw_exact)
graph.setLegends(["Instrumental PDF", "exact PDF"])
graph.setLegendPosition("upper right")
graph.setTitle("Independent Metropolis-Hastings: Instrumental PDF")
graph.setXTitle("x")
_ = otv.View(graph)

# %%
# :class:`~openturns.MetropolisHastings` and derived classes can work directly with the logarithm of the unscaled
# target density.

log_density = ot.ComposedFunction(ot.SymbolicFunction("x", "log(x)"), f)

initial_State = ot.Point([3.0])  # not important in this case
independent_IMH = ot.IndependentMetropolisHastings(
    log_density, range_PDF, initial_State, instrumental_dist, [0]
)

# %%
# Get a sample
sample_Size = 500
sample_IMH = independent_IMH.getSample(sample_Size)


# %%
# Plot the PDF of the sample to compare it to the target density
ks_algo = ot.KernelSmoothing()
ks_algo.setBoundaryCorrection(True)
ks_algo.setLowerBound(0.0)
ks_algo.setUpperBound(2.0 * pi)
posterior_IMH = ks_algo.build(sample_IMH)

g_IMH = posterior_IMH.drawPDF(-0.5, upper_bound * 1.1, 1001)
g_IMH.add(draw_exact)
g_IMH.setLegends(["IMH size = {}".format(sample_Size), "exact PDF"])
g_IMH.setTitle("Independent Metropolis-Hastings (IMH)")
g_IMH.setXTitle("x")
g_IMH.setLegendPosition("upper right")
view = otv.View(g_IMH)

# %%
# Even with very few sampling points (500),
# independent Metropolis-Hastings
# (with a judiciously chosen instrumental distribution)
# manages to capture the main features of the target density.

# %%
# Case 3(b): Use of Random walk Metropolis-Hastings
# -------------------------------------------------
# Let us use a normal instrumental distribution.

instrumental_dist = ot.Normal(0.0, 2.5)
randomwalk_MH = ot.RandomWalkMetropolisHastings(
    log_density, range_PDF, initial_State, instrumental_dist, [0]
)

# %%
# Get a sample
sample_RWMH = randomwalk_MH.getSample(sample_Size)

# %%
# Plot the PDF of the sample to compare it to the target density

ks_algo = ot.KernelSmoothing()
ks_algo.setBoundaryCorrection(True)
ks_algo.setLowerBound(0.0)
ks_algo.setUpperBound(2.0 * pi)
posterior_RWMH = ks_algo.build(sample_RWMH)

g_RWMH = posterior_RWMH.drawPDF(-0.5, upper_bound * 1.1, 1001)
g_RWMH.add(draw_exact)
g_RWMH.setLegends(["RWMH size = {}".format(sample_Size), "exact PDF"])
g_RWMH.setTitle("Random walk Metropolis-Hastings (RWMH)")
g_RWMH.setXTitle("x")
g_RWMH.setLegendPosition("upper right")
view = otv.View(g_RWMH)

# %%
# Final comparison
# ----------------
# We plot on the same graph the estimated PDF with all the previous algorithms with the same sample size.
# sphinx_gallery_thumbnail_number = 8
g = ks_pdf_GS.drawPDF(-0.5, upper_bound * 1.1, 1001)
g.add(posterior_RWMH.drawPDF(-0.5, upper_bound * 1.1, 1001))
g.add(posterior_IMH.drawPDF(-0.5, upper_bound * 1.1, 1001))
g.add(ks_pdf_RoU.drawPDF(-0.5, upper_bound * 1.1, 1001))
draw_exact.setLineStyle("dashed")
draw_exact.setColor("black")
g.add(draw_exact)
g.setLegends(["GS", "RWMH", "IMH", "RoU", "exact PDF"])
g.setTitle("Comparison of samplers (size = 500)")
g.setXTitle("x")
view = otv.View(g)


# %%
view.ShowAll()

# %%
# References
# -----------
# [1] Marin , J.M. & Robert, C.P. (2007). *Bayesian Core: A Practical Approach to Computational Bayesian Statistics*. Springer-Verlag, New York



#! /usr/bin/env python

import openturns as ot

dist1 = ot.Normal(1.0, 0.5)
print("dist1:", dist1)
result = dist1 + 2.0
print("dist1+2:", result)
graph = result.drawPDF()

result = dist1 - 2.0
print("dist1-2:", result)
graph = result.drawPDF()

result = dist1 * 2.0
print("dist1*2:", result)
graph = result.drawPDF()

result = dist1 / 2.0
print("dist1/2:", result)
graph = result.drawPDF()

result = 2.0 / dist1
print("2/dist1:", result)

result = dist1.cos()
print("cos(dist1):", result)
graph = result.drawPDF()

result = dist1.sin()
print("sin(dist1):", result)
graph = result.drawPDF()

result = dist1.tan()
print("tan(dist1):", result)
# graph = result.drawPDF()


dist0 = ot.Uniform(-0.999, 0.999)
result = dist0.acos()
print("acos(dist0):", result)
graph = result.drawPDF()

result = dist0.asin()
print("asin(dist0):", result)
graph = result.drawPDF()

result = dist0.atan()
print("atan(dist0):", result)
graph = result.drawPDF()


result = dist1.cosh()
print("cosh(dist1):", result)
graph = result.drawPDF()

result = dist1.sinh()
print("sinh(dist1):", result)
graph = result.drawPDF()

result = dist1.tanh()
print("tanh(dist1):", result)
graph = result.drawPDF()


distG1 = ot.LogNormal(1.0, 1.0, 1.0)
result = distG1.acosh()
print("acosh(distG1):", result)
graph = result.drawPDF()

result = dist1.asinh()
print("asinh(dist1):", result)
graph = result.drawPDF()

result = dist0.atanh()
print("atanh(dist0):", result)
graph = result.drawPDF()


result = dist1.exp()
print("exp(dist1):", result)
graph = result.drawPDF()

result = distG1.log()
print("log(distG1):", result)
graph = result.drawPDF()

result = distG1.ln()
print("ln(distG1):", result)
graph = result.drawPDF()


result = dist1**3
print("dist1^3:", result)
graph = result.drawPDF()

result = distG1**2.5
print("dist1^2.5:", result)
graph = result.drawPDF()

result = distG1.inverse()
print("inverse(distG1):", result)
graph = result.drawPDF()

result = dist1.sqr()
print("sqr(dist1):", result)
graph = result.drawPDF()

result = distG1.sqrt()
print("sqrt(distG1):", result)
graph = result.drawPDF()

result = dist1.cbrt()
print("cbrt(dist1):", result)
graph = result.drawPDF()

result = dist1.abs()
print("abs(dist1):", result)
graph = result.drawPDF()


dist2 = ot.Normal(-2.0, 1.0)
result = dist1 + dist2
print("dist1+dist2:", result)
graph = result.drawPDF()

result = dist1 - dist2
print("dist1-dist2:", result)
graph = result.drawPDF()

result = dist1 * dist2
print("dist1*dist2:", result)
graph = result.drawPDF()

result = dist1 / dist2
print("dist1/dist2:", result)
# graph = result.drawPDF()

result = 3 / dist1**2
print("3/dist1^2:", result)
graph = result.drawPDF()

result = (3 / dist1) ** 2
print("(3/dist1)^2:", result)
graph = result.drawPDF()

result = ot.LogNormal() * ot.LogNormal()
print("logn*logn:", result)
graph = result.drawPDF()

result = ot.LogUniform() * ot.LogUniform()
print("logu*logu:", result)
graph = result.drawPDF()

result = ot.LogUniform() * ot.LogNormal()
print("logu*logn:", result)
graph = result.drawPDF()

result = ot.LogNormal() * ot.LogUniform()
print("logn*logu:", result)
graph = result.drawPDF()

# For ticket #917
result = ot.WeibullMin() + ot.Exponential()
print("WeibullMin+Exponential:", result)
print("result.CDF(1.0)=%.6f" % result.computeCDF(1.0))
result = -1.0 * ot.WeibullMin() + ot.Exponential()
print("-WeibullMin+Exponential:", result)
print("result.CDF(1.0)=%.6f" % result.computeCDF(1.0))
result = ot.WeibullMin() - ot.Exponential()
print("WeibullMin-Exponential:", result)
print("result.CDF(1.0)=%.6f" % result.computeCDF(1.0))
result = -1.0 * ot.WeibullMin() - ot.Exponential()
print("-WeibullMin-Exponential:", result)
print("result.CDF(-1.0)=%.6f" % result.computeCDF(-1.0))

# 2-d
print(ot.Normal(2) + ot.Normal(2))
print(ot.Normal(2) + 3.0)
print(ot.Normal(2) - ot.Normal(2))
print(ot.Normal(2) - 3.0)

# unary minus
x = ot.Normal(7.0, 2.0)
print(-x)
print(-ot.Distribution(x))

# simplification of sum
x = -ot.Exponential() - ot.Exponential()
print(x)

# take into account the weight and the constant in simplification
x = 2 * (-ot.Exponential() - ot.Exponential())
print(x)
x = 2 * (-ot.Exponential() - ot.Exponential()) + 1.0
print(x)
x = ot.Poisson(5.0) + 1.0
print(x)



#! /usr/bin/env python

import openturns as ot
import numpy as np

np.set_printoptions(
    formatter={"float": "{: 4.1f}".format, "complexfloat": "{: 4.1f}".format}
)

ot.TESTPREAMBLE()

# Check tuple / Point conversion
t0 = (0.5, 1.5)
p0 = ot.Point(t0)
print("tuple", t0, "=> Point", p0)

t1 = tuple(p0)
print("Point", p0, "=> tuple", t1)

print("Point", p0, "+ tuple", t0, "=> Point", p0 + t0)

# Check list / Point conversion
l0 = [0.5, 1.5]
p0 = ot.Point(l0)
print("list", l0, "=> Point", p0)

l1 = list(p0)
print("Point", p0, "=> list", l1)

print("Point", p0, "+ list", l0, "=> Point", p0 + l0)

# Check sequence protocol for Point
for x in p0:
    print("value", x)
x0 = p0[0]
p0[0] = x0
if x0 not in p0:
    raise ValueError("Point badly implements __contains__()")

# Check array / Point conversion
a0 = np.array((0.5, 1.5))
p0 = ot.Point(a0)
print("array", a0, "=> Point", p0)

a1 = np.array(p0)
print("Point", p0, "=> array", a1)

print("Point", p0, "+ array", a0, "=> Point", p0 + a0)
print("array", a0, "+ Point", p0, "=> array", a0 + p0)

# See ticket #423
m0 = np.array([[1, 1], [2, 2], [3, 3]])
try:
    p0 = ot.Point(m0)
    print(p0)
except Exception:
    print("Conversion from 2d-array => Point failed (as expected :)")

# Check tuple / Sample conversion
t0 = ((1.0, 2.0), (3.0, 4.0))
s0 = ot.Sample(t0)
print("tuple", t0, "=> Sample", s0)

t0 = ([1.0, 2.0], [3.0, 4.0])
s0 = ot.Sample(t0)
print("tuple", t0, "=> Sample", s0)

t1 = tuple(s0)
print("Sample", s0, "=> tuple", t1)

# Check list / Sample conversion
l0 = [[1.0, 2.0], [3.0, 4.0]]
s0 = ot.Sample(l0)
print("list", l0, "=> Sample", s0)

l0 = [(1.0, 2.0), (3.0, 4.0)]
s0 = ot.Sample(l0)
print("list", l0, "=> Sample", s0)

l1 = list(s0)
print("Sample", s0, "=> list", l1)

# Check array / Sample conversion
a0 = np.array(((1.0, 2.0), (3.0, 4.0)))
s0 = ot.Sample(a0)
print("array", a0, "=> Sample", s0)

a1 = np.array(s0)
print("Sample", s0, "=> array", a1)

s0 = ot.Sample(a0.transpose())
print("with transpose, array", a0.transpose(), "=> Sample", s0)

# Check tuple / Function interoperability
F = ot.SymbolicFunction(("E", "F", "L", "I"), ("-F*L^3/(3.*E*I)",))

t0 = (1.0, 2.0, 3.0, 4.0)
print("Point", F(t0), "= F( tuple", t0, ")")

# Check list / Function interoperability
l0 = [1.0, 2.0, 3.0, 4.0]
print("Point", F(l0), "= F( list", l0, ")")

# Check array / Function interoperability
a0 = np.array((1.0, 2.0, 3.0, 4.0))
print("Point", F(a0), "= F( array", a0, ")")
a1 = np.array(((1.0, 2.0, 3.0, 4.0), (5.0, 6.0, 7.0, 8.0)))
print("Sample", F(a1), "= F( array", a1, ")")

# Check Python function / Function interoperability


def aFunc(x):
    return [x[0] + x[1] + x[2] + x[3]]


PYNMF = ot.PythonFunction(4, 1, aFunc)

print("Point", PYNMF(t0), "= PYNMF( tuple", t0, ")")
print("Point", PYNMF(l0), "= PYNMF( list", l0, ")")
print("Point", PYNMF(a0), "= PYNMF( array", a0, ")")
print("Sample", PYNMF(a1), "= PYNMF( array", a1, ")")

# Check 2-d array which nested dim is size=1 / Function
# interoperability


def aFunc2(x):
    return [2.0 * x[0]]


PYNMF = ot.PythonFunction(1, 1, aFunc2)

a0 = np.array(([1.0]))
print("Point", PYNMF(a0), "= PYNMF( array", a0, ")")
a1 = np.array(([1.0], [2.0], [3.0]))
print("Sample", PYNMF(a1), "= PYNMF( array", a1, ")")

# Check tuple / Indices conversion
t0 = (1, 2)
i0 = ot.Indices(t0)
print("tuple", t0, "=> Indices", i0)

t1 = tuple(i0)
print("Indices", i0, "=> tuple", tuple([int(x) for x in t0]))

# Check list / Indices conversion
l0 = [3, 4, 5]
i0 = ot.Indices(l0)
print("list", l0, "=> Indices", i0)

l1 = list(i0)
print("Indices", i0, "=> list", [int(x) for x in l1])

# check Indices typemap
sample = ot.Normal(3).getSample(10)
print(sample.getDescription())
marginal = sample.getMarginal((0, 2))
print(marginal.getDescription())
marginal = sample.getMarginal([0, 2])
print(marginal.getDescription())

# check array / IndicesCollection conversion
a0 = np.array(((1, 3, 5), (2, 4, 6)))
i0 = ot.IndicesCollection(a0)
print("array", a0, "=> IndicesCollection", i0)

a1 = np.array(i0)
print("IndicesCollection", i0, "=> array", a1)

# Check sequence protocol for IndicesCollection
for x in i0:
    print("value", x)

# Check tuple / Description conversion
t0 = ("blob", "zou")
i0 = ot.Description(t0)
print("tuple", t0, "=> Description", i0)

t1 = tuple(i0)
print("Description", i0, "=> tuple", t1)

# Check list / Description conversion
l0 = ["blob", "zou"]
i0 = ot.Description(l0)
print("list", l0, "=> Description", i0)

l1 = list(i0)
print("Description", i0, "=> list", l1)

# array / Description conversion
a0 = np.array(("x0", "x1", "x2"))
d0 = ot.Description(a0)
print("array", a0, "=> Description", d0)

a1 = np.array(d0)
print("Description", d0, "=> array", a1)

# check Description typemap
sample.setDescription(("x0", "x1", "x2"))
print(sample.getDescription())
sample.setDescription(("y0", "y1", "y2"))
print(sample.getDescription())
sample.setDescription(np.array(("z0", "z1", "z2")))
print(sample.getDescription())

# Check Matrix tuple constructor
t0 = (1.0, 2.0, 3.0, 4.0)

m0 = ot.Matrix(2, 2, t0)
print("tuple", t0, "=> Matrix", m0)

m0 = ot.SquareMatrix(2, t0)
print("tuple", t0, "=> SquareMatrix", m0)

m0 = ot.SymmetricMatrix(2, t0)
print("tuple", t0, "=> SymmetricMatrix", m0)

m0 = ot.Tensor(2, 2, 1, t0)
print("tuple", t0, "=> Tensor", m0)

m0 = ot.SymmetricTensor(2, 1, t0)
print("tuple", t0, "=> SymmetricTensor", m0)

m0 = ot.CorrelationMatrix(2, t0)
print("tuple", t0, "=> CorrelationMatrix", m0)

m0 = ot.CovarianceMatrix(2, t0)
print("tuple", t0, "=> CovarianceMatrix", m0)

t0c = (1.0 + 3.0j, 2.0 - 5.0j, 3.0 + 7.0j, 4.0 - 9.0j)

m0 = ot.ComplexMatrix(2, 2, t0)
print("tuple", t0, "=> ComplexMatrix", m0)

m0 = ot.ComplexMatrix(2, 2, t0c)
print("tuple", t0c, "=> ComplexMatrix", m0)

m0 = ot.SquareComplexMatrix(2, t0)
print("tuple", t0, "=> SquareComplexMatrix", m0)

m0 = ot.SquareComplexMatrix(2, t0c)
print("tuple", t0c, "=> SquareComplexMatrix", m0)

# Check Matrix list constructor
l0 = [1.0, 2.0, 3.0, 4.0]
m0 = ot.Matrix(2, 2, l0)
print("list", l0, "=> Matrix", m0)

m0 = ot.SquareMatrix(2, l0)
print("list", l0, "=> SquareMatrix", m0)

m0 = ot.SymmetricMatrix(2, l0)
print("list", l0, "=> SymmetricMatrix", m0)

m0 = ot.Tensor(2, 2, 1, l0)
print("list", l0, "=> Tensor", m0)

m0 = ot.SymmetricTensor(2, 1, l0)
print("list", l0, "=> SymmetricTensor", m0)

m0 = ot.CorrelationMatrix(2, l0)
print("list", l0, "=> CorrelationMatrix", m0)

m0 = ot.CovarianceMatrix(2, l0)
print("list", l0, "=> CovarianceMatrix", m0)

l0c = [1.0 + 3.0j, 2.0 - 5.0j, 3.0 + 7.0j, 4.0 - 9.0j]

m0 = ot.ComplexMatrix(2, 2, l0)
print("list", l0, "=> ComplexMatrix", m0)

m0 = ot.ComplexMatrix(2, 2, l0c)
print("list", l0c, "=> ComplexMatrix", m0)

m0 = ot.SquareComplexMatrix(2, l0)
print("list", l0, "=> SquareComplexMatrix", m0)

m0 = ot.SquareComplexMatrix(2, l0c)
print("list", l0c, "=> SquareComplexMatrix", m0)

# Check Matrix 1-d array constructor
a0 = np.array((1.0, 2.0, 3.0, 4.0))
m0 = ot.Matrix(2, 2, a0)
print("array", a0, "=> Matrix", m0)

m0 = ot.SquareMatrix(2, a0)
print("array", a0, "=> SquareMatrix", m0)

m0 = ot.SymmetricMatrix(2, a0)
print("array", a0, "=> SymmetricMatrix", m0)

m0 = ot.Tensor(2, 2, 1, a0)
print("array", a0, "=> Tensor", m0)

m0 = ot.SymmetricTensor(2, 1, a0)
print("array", a0, "=> SymmetricTensor", m0)

m0 = ot.CorrelationMatrix(2, a0)
print("array", a0, "=> CorrelationMatrix", m0)

m0 = ot.CovarianceMatrix(2, a0)
print("array", a0, "=> CovarianceMatrix", m0)

a0c = np.array((1.0 + 3j, 2.0 - 5j, 3.0 - 7j, 4.0 + 9j))

m0 = ot.ComplexMatrix(2, 2, a0)
print("array", a0, "=> ComplexMatrix", m0)

m0 = ot.ComplexMatrix(2, 2, a0c)
print("array", a0c, "=> ComplexMatrix", m0)

m0 = ot.SquareComplexMatrix(2, a0)
print("array", a0, "=> SquareComplexMatrix", m0)

m0 = ot.SquareComplexMatrix(2, a0c)
print("array", a0c, "=> SquareComplexMatrix", m0)

# check array / Matrix conversion
a0 = np.array(((1.0, 2.0), (3.0, 4.0), (5.0, 6.0)))
m0 = ot.Matrix(a0)
print("array", a0, "=> Matrix", m0)
a1 = np.array(m0)
print("Matrix", m0, "=> array", a1)
m0 = ot.Matrix(a0.transpose())
print("with transpose, array", a0.transpose(), "=> Matrix", m0)

a0 = np.array(((1.0, 2.0), (3.0, 4.0)))
m0 = ot.SquareMatrix(a0)
print("array", a0, "=> SquareMatrix", m0)
a1 = np.array(m0)
print("SquareMatrix", m0, "=> array", a1)

a0 = np.array(((1.0, 2.0), (0.0, 4.0)))
m0 = ot.TriangularMatrix(a0)
print("array", a0, "=> TriangularMatrix", m0)
a1 = np.array(m0)
print("TriangularMatrix", m0, "=> array", a1)

a0 = np.array(((1.0, 2.0), (2.0, 4.0)))
m0 = ot.SymmetricMatrix(a0)
print("array", a0, "=> SymmetricMatrix", m0)
m0[1, 0] = 3.0
a1 = np.array(m0)
print("SymmetricMatrix", m0, "=> array", a1)

a0 = np.array(
    (
        ((1.0, 2.0), (3.0, 4.0), (5.0, 6.0)),
        ((7.0, 8.0), (9.0, 10.0), (11.0, 12.0)),
        ((13.0, 14.0), (15.0, 16.0), (17.0, 18.0)),
        ((19.0, 20.0), (21.0, 22.0), (23.0, 24.0)),
    )
)
m0 = ot.Tensor(a0)
print("array", a0, "=> Tensor", m0)
a1 = np.array(m0)
print("Tensor", m0, "=> array", a1)
m0 = ot.Tensor(a0.transpose())
print("with transpose, array", a0.transpose(), "=> Tensor", m0)

a0 = np.array((((1.0, 2.0), (3.0, 4.0)), ((3.0, 4.0), (7.0, 8.0))))
m0 = ot.SymmetricTensor(a0)
print("array", a0, "=> SymmetricTensor", m0)
m0[1, 0, 0] = 9.0
a1 = np.array(m0)
print("SymmetricTensor", m0, "=> array", a1)

a0 = np.array(((2.0, 1.0), (1.0, 2.0)))
m0 = ot.CovarianceMatrix(a0)
print("array", a0, "=> CovarianceMatrix", m0)
a1 = np.array(m0)
print("CovarianceMatrix", m0, "=> array", a1)

a0 = np.array(((1.0, 0.5), (0.5, 1.0)))
m0 = ot.CorrelationMatrix(a0)
print("array", a0, "=> CorrelationMatrix", m0)
a1 = np.array(m0)
print("CorrelationMatrix", m0, "=> array", a1)

a0 = np.array(((1.0 + 3j, 2.0 - 5j, 3.0 + 7j), (4.0 - 9j, 5.0 + 11j, 6.0 - 13j)))
m0 = ot.ComplexMatrix(a0)
print("array", a0, "=> ComplexMatrix", m0)
a1 = np.array(m0)
print("ComplexMatrix", m0, "=> array", a1)
m0 = ot.ComplexMatrix(a0.transpose())
print("with transpose, array", a0.transpose(), "=> ComplexMatrix", m0)

a0 = np.array(((1.0 + 3j, 2.0 - 5j), (3.0 + 7j, 4.0 - 9j)))
m0 = ot.SquareComplexMatrix(a0)
print("array", a0, "=> SquareComplexMatrix", m0)
a1 = np.array(m0)
print("SquareComplexMatrix", m0, "=> array", a1)

a0 = np.array(((1.0 + 3j, 0.0), (3.0 + 7j, 4.0 - 9j)))
m0 = ot.TriangularComplexMatrix(a0)
print("array", a0, "=> TriangularComplexMatrix", m0)
a1 = np.array(m0)
print("TriangularComplexMatrix", m0, "=> array", a1)

a0 = np.array(((1.0 + 3j, 2.0 - 5j), (2.0 + 5j, 4.0 - 9j)))
m0 = ot.HermitianMatrix(a0)
print("array", a0, "=> HermitianMatrix", m0)
m0[1, 0] = 3.0 - 5j
a1 = np.array(m0)
print("HermitianMatrix", m0, "=> array", a1)

# check np.matrix / Matrix conversion
a0 = np.matrix(((1.0, 2.0), (3.0, 4.0), (5.0, 6.0)))
m0 = ot.Matrix(a0)
print("matrix", a0, "=> Matrix", m0)
a1 = np.array(m0)
print("Matrix", m0, "=> matrix", a1)
m0 = ot.Matrix(a0.transpose())
print("with transpose, matrix", a0.transpose(), "=> Matrix", m0)

a0 = np.matrix(((1.0 + 3j, 2.0 - 5j, 3.0 + 7j), (4.0 - 9j, 5.0 + 11j, 6.0 - 13j)))
m0 = ot.ComplexMatrix(a0)
print("matrix", a0, "=> ComplexMatrix", m0)
a1 = np.array(m0)
print("ComplexMatrix", m0, "=> matrix", a1)
m0 = ot.ComplexMatrix(a0.transpose())
print("with transpose, matrix", a0.transpose(), "=> ComplexMatrix", m0)

# empty array
a0 = np.zeros((0, 3))
s0 = ot.Sample(a0)
print("empty array => sample", repr(s0))
a1 = np.array(s0)
print("sample => array", a1, a1.shape)

# Sample int64 indexing
s0 = ot.Sample(5, 3)
idx = np.int64(2)
print("sample[int64]:", s0[idx])
s0[idx] = [42] * 3
print("sample[int64]=Point:", s0)

# Sample [int64] indexing
s0 = ot.Sample(5, 3)
idx = np.array([1, 3, 4])
print("sample[[int64]]:", s0[idx])
s0[idx] = ot.Normal(3).getSample(3)
print("sample[[int64]]=Sample:", s0)

# generic int64 indexing
s0 = ot.Description(5, "aa")
idx = np.int64(2)
print("Description[int64]:", s0[idx])
s0[idx] = "zou"
print("Description[int64]=str", s0[idx])

# generic [int64] indexing
s0 = ot.Description(5, "aa")
idx = np.array([1, 3, 4])
print("Description[[int64]]:", s0[idx])
s0[idx] = ["zou"] * 3
print("Description[[int64]]=str", s0[idx])

# Field int64 indexing
mesher = ot.IntervalMesher([10, 5])
mesh = mesher.build(ot.Interval([0.0, 0.0], [2.0, 1.0]))
process = ot.GaussianProcess(ot.ExponentialModel([0.2] * 2, [1.0]), mesh)
field = process.getRealization()
idx = np.int64(2)
print("Field[int64]", field[idx])
field[idx] = [6.0]
print("Field[int64]=Point", field[idx])



"""
Estimate a conditional quantile
===============================
"""

# sphinx_gallery_thumbnail_number = 8
# %%
# From a multivariate data sample, we estimate a distribution with kernel smoothing.
# Here we present a bivariate distribution :math:`X= (X_1, X_2)`.
# We use the `computeConditionalQuantile` method to estimate the 90% quantile :math:`Q_1` of the conditional variable :math:`X_2|X_1` :
#
# .. math::
#    Q_2 : x_1 \mapsto q_{0.9}(X_2|X_1=x_1)
#
# We then draw the curve :math:`Q_2 : x_1 \mapsto Q_2(x_1)`.
# We first start with independent Normals then we consider dependent marginals with a Clayton copula.
#

# %%
import openturns as ot
import openturns.viewer as otv
import numpy as np

# %%
# Defining the marginals
# ----------------------
#
# We consider two independent Normal marginals :
X1 = ot.Normal(0.0, 1.0)
X2 = ot.Normal(0.0, 3.0)


# %%
# Independent marginals
# ---------------------

distX = ot.JointDistribution([X1, X2])
sample = distX.getSample(1000)


# %%
# Let's see the data

# %%
graph = ot.Graph("2D-Normal sample", "x1", "x2")
cloud = ot.Cloud(sample, "blue", "fsquare", "My Cloud")
graph.add(cloud)
graph.setXTitle("$X_1$")
graph.setYTitle("$X_2$")
graph.setTitle("A sample from $X=(X_1, X_2)$")
view = otv.View(graph)

# %%
# We draw the isolines of the PDF of :math:`X` :
graph = distX.drawPDF()
graph.setXTitle("$X_1$")
graph.setYTitle("$X_2$")
graph.setTitle("iso-PDF of $X=(X_1, X_2)$")
view = otv.View(graph)

# %%
# We estimate the density with kernel smoothing :
kernel = ot.KernelSmoothing()
estimated = kernel.build(sample)


# %%
# We draw the isolines of the estimated PDF of :math:`X` :
graph = estimated.drawPDF()
graph.setXTitle("$X_1$")
graph.setYTitle("$X_2$")
graph.setTitle("iso-PDF of $X=(X_1, X_2)$ estimated by kernel smoothing")
view = otv.View(graph)


# %%
# We can compute the conditional quantile of :math:`X_2 | X_1` with the `computeConditionalQuantile` method and draw it after.
#

# %%
# We first create :math:`\sampleSize` observation points in :math:`[-3.0, 3.0]` :

# %%
N = 301
xobs = np.linspace(-3.0, 3.0, N)
sampleObs = ot.Sample([[xi] for xi in xobs])

# %%
# We create curves of the exact and approximated quantile :math:`Q_1`

# %%
x = [xi for xi in xobs]
yapp = [estimated.computeConditionalQuantile(0.9, sampleObs[i]) for i in range(N)]
yex = [distX.computeConditionalQuantile(0.9, sampleObs[i]) for i in range(N)]

# %%
cxy_app = ot.Curve(x, yapp)
cxy_ex = ot.Curve(x, yex)
graph = ot.Graph("90% quantile of $X_2 | X_1=x_1$", "$x_1$", "$Q_2(x_1)$")
graph.add(cxy_app)
graph.add(cxy_ex)
graph.setLegends(["$Q_2$ kernel smoothing", "$Q_2$ exact"])
graph.setLegendPosition("lower right")
view = otv.View(graph)

# %%
# In this case the :math:`Q_2` quantile is constant because of the independence of the marginals.
#

# %%
# Dependence through a Clayton copula
# -----------------------------------
#
# We now define a Clayton copula to model the dependence between our marginals.
# The Clayton copula is a bivariate asymmmetric Archimedean copula, exhibiting greater dependence
# in the negative tail than in the positive.
#

# %%
copula = ot.ClaytonCopula(2.5)
distX = ot.JointDistribution([X1, X2], copula)

# %%
# We generate a sample from the distribution :
sample = distX.getSample(1000)

# %%
# Let's see the data

# %%
graph = ot.Graph("2D-Normal sample", "x1", "x2")
cloud = ot.Cloud(sample, "blue", "fsquare", "My Cloud")
graph.add(cloud)
graph.setXTitle("$X_1$")
graph.setYTitle("$X_2$")
graph.setTitle("A sample from $X=(X_1, X_2)$")
view = otv.View(graph)

# %%
# We draw the isolines of the PDF of :math:`X` :
graph = distX.drawPDF()
graph.setXTitle("$X_1$")
graph.setYTitle("$X_2$")
graph.setTitle("iso-PDF of $X=(X_1, X_2)$")
view = otv.View(graph)

# %%
# We estimate the density with kernel smoothing :
kernel = ot.KernelSmoothing()
estimated = kernel.build(sample)

# %%
# We draw the isolines of the estimated PDF of :math:`X` :
graph = estimated.drawPDF()
graph.setXTitle("$X_1$")
graph.setYTitle("$X_2$")
graph.setTitle("iso-PDF of $X=(X_1, X_2)$ estimated by kernel smoothing")
view = otv.View(graph)


# %%
# We can compute the conditional quantile of :math:`X_2 | X_1=x1` with the `computeConditionalQuantile` method and draw it after.
#

# %%
# We first create :math:`\sampleSize` observation points in :math:`[-3.0, 3.0]` :

# %%
N = 301
xobs = np.linspace(-3.0, 3.0, N)
sampleObs = ot.Sample([[xi] for xi in xobs])

# %%
# We create curves of the exact and approximated quantile :math:`Q_1`

# %%
x = [xi for xi in xobs]
yapp = [estimated.computeConditionalQuantile(0.9, sampleObs[i]) for i in range(N)]
yex = [distX.computeConditionalQuantile(0.9, sampleObs[i]) for i in range(N)]

# %%
cxy_app = ot.Curve(x, yapp)
cxy_ex = ot.Curve(x, yex)
graph = ot.Graph("90% quantile of $X_2 | X_1=x_1$", "$x_1$", "$Q_2(x_1)$")
graph.add(cxy_app)
graph.add(cxy_ex)
graph.setLegends(["$Q_2$ kernel smoothing", "$Q_2$ exact"])
graph.setLegendPosition("lower right")
view = otv.View(graph)

# %%
# Our estimated conditional quantile is a good approximate and should be better the more data we have. We can observe it by increasing the number of samples.

# %%
# Display the graphs
view.ShowAll()



import openturns as ot
import openturns.testing as ott
from openturns.usecases.fire_satellite import FireSatelliteModel

ot.TESTPREAMBLE()


def use_case_1(X, Y):
    """
    optim problem (scale)
    Dirac model
    """
    ot.ResourceMap.Reset()
    basis = ot.LinearBasisFactory(inputDimension).build()
    # Case of a misspecified covariance model
    covarianceModel = ot.DiracCovarianceModel(inputDimension)
    algo = ot.GaussianProcessFitter(X, Y, covarianceModel, basis)
    assert algo.getOptimizeParameters()
    assert algo.getKeepCholeskyFactor()
    algo.setKeepCholeskyFactor(False)
    algo.run()
    cov_amplitude = [0.19575]
    trend_coefficients = [-0.1109, 1.0149]
    result = algo.getResult()
    ott.assert_almost_equal(
        result.getCovarianceModel().getAmplitude(), cov_amplitude, 1e-4, 1e-4
    )
    ott.assert_almost_equal(
        result.getTrendCoefficients(), trend_coefficients, 1e-4, 1e-4
    )


def use_case_2(X, Y):
    """
    No optim with Dirac model
    """
    ot.ResourceMap.Reset()
    basis = ot.LinearBasisFactory(inputDimension).build()
    # Case of a misspecified covariance model
    covarianceModel = ot.DiracCovarianceModel(inputDimension)
    algo = ot.GaussianProcessFitter(X, Y, covarianceModel, basis)
    assert algo.getKeepCholeskyFactor()
    algo.setKeepCholeskyFactor(False)
    algo.setOptimizeParameters(False)
    algo.run()
    cov_amplitude = [1]
    trend_coefficients = [-0.1109, 1.01498]
    result = algo.getResult()
    ott.assert_almost_equal(
        result.getCovarianceModel().getAmplitude(), cov_amplitude, 1e-4, 1e-4
    )
    ott.assert_almost_equal(
        result.getTrendCoefficients(), trend_coefficients, 1e-4, 1e-4
    )


def use_case_3(X, Y):
    """
    full optim problem (scale)
    analytical variance estimate
    """
    ot.ResourceMap.Reset()
    basis = ot.LinearBasisFactory(inputDimension).build()
    # Case of a misspecified covariance model
    covarianceModel = ot.AbsoluteExponential(inputDimension)
    algo = ot.GaussianProcessFitter(X, Y, covarianceModel, basis)
    assert algo.getOptimizeParameters()
    algo.setKeepCholeskyFactor(False)
    algo.run()
    cov_param = [0.0078, 1]
    trend_coefficients = [-0.110943, 1.01498]
    result = algo.getResult()
    assert (
        algo.getOptimizationAlgorithm().getImplementation().getClassName() == "Cobyla"
    )
    print(result.getCovarianceModel().getParameter())
    ott.assert_almost_equal(
        result.getCovarianceModel().getParameter(), cov_param, 1e-4, 1e-4
    )
    print(result.getTrendCoefficients())
    ott.assert_almost_equal(
        result.getTrendCoefficients(), trend_coefficients, 1e-4, 1e-4
    )


def use_case_4(X, Y):
    """
    optim problem (scale)
    Biased variance estimate
    """
    ot.ResourceMap.Reset()
    ot.ResourceMap.SetAsBool("GeneralLinearModelAlgorithm-UnbiasedVariance", False)
    basis = ot.LinearBasisFactory(inputDimension).build()
    # Case of a misspecified covariance model
    covarianceModel = ot.AbsoluteExponential(inputDimension)
    algo = ot.GaussianProcessFitter(X, Y, covarianceModel, basis)
    assert algo.getKeepCholeskyFactor()
    assert algo.getOptimizeParameters()
    algo.setKeepCholeskyFactor(False)
    algo.run()
    result = algo.getResult()
    assert (
        algo.getOptimizationAlgorithm().getImplementation().getClassName() == "Cobyla"
    )
    print(result.getCovarianceModel().getParameter())
    ott.assert_almost_equal(
        result.getCovarianceModel().getParameter(), [0.0078, 1], 1e-4, 1e-4
    )
    print(result.getTrendCoefficients())
    ott.assert_almost_equal(
        result.getTrendCoefficients(), [-0.110943, 1.01498], 1e-4, 1e-4
    )


def use_case_5(X, Y):
    """
    full optim problem (scale, amplitude)
    """
    ot.ResourceMap.Reset()
    ot.ResourceMap.SetAsBool("GaussianProcessFitter-UnbiasedVariance", False)
    ot.ResourceMap.SetAsBool(
        "GaussianProcessFitter-UseAnalyticalAmplitudeEstimate", False
    )
    basis = ot.LinearBasisFactory(inputDimension).build()
    # Case of a misspecified covariance model
    covarianceModel = ot.AbsoluteExponential(inputDimension)

    algo = ot.GaussianProcessFitter(X, Y, covarianceModel, basis)
    assert algo.getKeepCholeskyFactor()
    assert algo.getOptimizeParameters()
    algo.setKeepCholeskyFactor(False)
    bounds = ot.Interval([1e-2] * 2, [100] * 2)
    algo.setOptimizationBounds(bounds)
    algo.run()
    result = algo.getResult()
    assert (
        algo.getOptimizationAlgorithm().getImplementation().getClassName() == "Cobyla"
    )
    print(result.getCovarianceModel().getParameter())
    cov_param = [8.499, 0.9998]
    ott.assert_almost_equal(
        result.getCovarianceModel().getParameter(), cov_param, 1e-3, 1e-3
    )
    print(result.getTrendCoefficients())
    trend_coefficients = [0.09867, 0.995]
    ott.assert_almost_equal(
        result.getTrendCoefficients(), trend_coefficients, 1e-4, 1e-4
    )


def use_case_6(X, Y):
    ot.RandomGenerator.SetSeed(0)
    ot.ResourceMap.Reset()
    ot.ResourceMap.SetAsBool(
        "GaussianProcessFitter-UseAnalyticalAmplitudeEstimate", False
    )
    covarianceModel = ot.AbsoluteExponential()
    algo = ot.GaussianProcessFitter(X, Y, covarianceModel)
    assert algo.getKeepCholeskyFactor()
    algo.setKeepCholeskyFactor(False)
    assert algo.getOptimizeParameters()
    algo.run()
    result = algo.getResult()
    assert (
        algo.getOptimizationAlgorithm().getImplementation().getClassName() == "Cobyla"
    )
    print(result.getCovarianceModel().getParameter())
    cov_param = [2.55922, 1]
    ott.assert_almost_equal(
        result.getCovarianceModel().getParameter(), cov_param, 1e-4, 1e-4
    )
    ott.assert_almost_equal(result.getTrendCoefficients(), [])


def use_case_7(X, Y):
    ot.RandomGenerator.SetSeed(0)
    ot.ResourceMap.Reset()
    ot.ResourceMap.SetAsScalar(
        "GaussianProcessFitter-OptimizationLowerBoundScaleFactor", 0.0
    )
    covarianceModel = ot.AbsoluteExponential()
    with ott.assert_raises(TypeError):
        algo = ot.GaussianProcessFitter(X, Y, covarianceModel)
        algo.run()


def use_case_8(X, Y):
    ot.RandomGenerator.SetSeed(0)
    ot.ResourceMap.Reset()
    covarianceModel = ot.AbsoluteExponential()
    algo = ot.GaussianProcessFitter(X, Y, covarianceModel)
    algo.run()
    result = algo.getResult()
    scale0, amplitude0 = result.getCovarianceModel().getParameter()
    algo.setNoise([1.0] * len(X))
    algo.run()
    result = algo.getResult()
    scale1, amplitude1 = result.getCovarianceModel().getParameter()
    assert scale1 < scale0, "scale comparison"
    assert amplitude1 < amplitude0, "amplitude comparison"


def bugfix_optim_no_feasible():
    ot.RandomGenerator.SetSeed(0)
    ot.ResourceMap.Reset()

    m = FireSatelliteModel()
    model = m.model
    inputDistribution = m.distribution

    experiment = ot.LHSExperiment(inputDistribution, 10 * m.dim)
    inputTrainingSet = experiment.generate()
    outputTrainingSet = model(inputTrainingSet)
    linear_basis = ot.LinearBasisFactory(m.dim).build()
    basis = ot.Basis(
        [
            ot.AggregatedFunction([linear_basis.build(k)] * 3)
            for k in range(linear_basis.getSize())
        ]
    )

    myCov1 = ot.MaternModel([1.0] * m.dim, 2.5)
    myCov2 = ot.SquaredExponential([1.0] * m.dim)
    myCov3 = ot.MaternModel([1.0] * m.dim, 2.5)

    # optimal we should get after the optimization process
    optimal_cov_parameter = [
        7.874e06,
        1513,
        1413,
        43.96,
        5.509,
        4.625,
        5.159,
        0.6612,
        3.245,
        4.945,
        8.074,
        5.419,
    ]
    covarianceModel = ot.TensorizedCovarianceModel([myCov1, myCov2, myCov3])

    scaleOptimizationBounds = ot.Interval(
        [1.0e6, 1.0e3, 1.0e3, 1.0, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
        [2.0e7, 2.0e3, 2.0e3, 1e2, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0],
    )

    algo = ot.GaussianProcessFitter(
        inputTrainingSet, outputTrainingSet, covarianceModel, basis
    )
    algo.setOptimizationBounds(scaleOptimizationBounds)
    algo.setOptimizeParameters(True)
    algo.run()
    # Get result & residual
    result = algo.getResult()
    residual = result.getMetaModel()(inputTrainingSet) - outputTrainingSet
    # Define multivariate square function
    sqr_func = ot.SymbolicFunction(["x", "y", "z"], ["x*x", "y*y", "z*z"])
    # Squared residual
    squared_epsilon = sqr_func(residual).computeMean()
    ott.assert_almost_equal(
        result.getCovarianceModel().getParameter(), optimal_cov_parameter, 1e-1, 1e-3
    )
    ott.assert_almost_equal(squared_epsilon, [7.248e-06, 298.4, 0.9051], 5e-1, 1e-3)


if __name__ == "__main__":

    ot.RandomGenerator.SetSeed(0)
    sampleSize = 40
    inputDimension = 1

    # Create the function to estimate
    model = ot.SymbolicFunction(["x0"], ["x0"])

    X = ot.Sample(sampleSize, inputDimension)
    for i in range(sampleSize):
        X[i, 0] = 3.0 + (8.0 * i) / sampleSize
    Y = model(X)

    # Add a small noise to data
    Y += (
        ot.GaussianProcess(ot.AbsoluteExponential([0.1], [0.2]), ot.Mesh(X))
        .getRealization()
        .getValues()
    )
    use_case_1(X, Y)
    use_case_2(X, Y)
    use_case_3(X, Y)
    use_case_4(X, Y)
    use_case_5(X, Y)
    use_case_6(X, Y)
    use_case_7(X, Y)
    use_case_8(X, Y)
    # fix https://github.com/openturns/openturns/issues/2953
    bugfix_optim_no_feasible()



"""
=========================================
Get the gradient and Hessian as functions
=========================================
"""

# %%
# Context
# =======
#
# Let :math:`f : \mathbb{R}^\inputDim \mapsto \mathbb{R}^q` a function defined by
# its component functions :math:`f_i: \mathbb{R}^d \mapsto \mathbb{R}`. We want to get its gradient
# and Hessian as functions. The way to get it depends on the type of the function :math:`f`.
#
# We can do that:
#
# - Case 1: This case covers the :class:`~openturns.SymbolicFunction`,
# - Case 2: This case covers the general case.
#
import openturns as ot

# %%
# Case 1: The :class:`~openturns.SymbolicFunction`
# ================================================
#
# Here, the function :math:`f` is a :class:`~openturns.SymbolicFunction`. In this
# particular case, we have access to the analytical
# expression of the gradient and hessian functions.
#
# Let :math:`f : \Rset^2 \rightarrow \Rset^2` be defined by:
#
# .. math::
#    f(\vect{x}) = (f_1(x_1, x_2), f_2(x_1, x_2)) = (x_1^2 + 2x_2^2, x_1 + 3x_2^4)
#
# We want to define the gradient function :math:`\Rset^2 \rightarrow \Rset^4` defined by:
#
# .. math::
#    :label: gradientFct
#
#    \vect{x} = (x_1, x_2) \rightarrow & \left(\dfrac{\partial f_1 (\vect{x})}{\partial x_1},
#                     \dfrac{\partial f_1 (\vect{x})}{\partial x_2},
#                     \dfrac{\partial f_2 (\vect{x})}{\partial x_1},
#                     \dfrac{\partial f_2 (\vect{x})}{\partial x_2}\right)\\
#                & = (2x_1, 4x_2, 1, 12x_2^3)
#
# To get the gradient as a function, we first get the analytical expressions of the derivatives:
f = ot.SymbolicFunction(["x1", "x2"], ["x1^2 + 2 * x2^2", "x1 + 3 * x2^4"])
formula_f1_deriv_x1 = f.getGradient().getImplementation().getFormula(0, 0)
formula_f1_deriv_x2 = f.getGradient().getImplementation().getFormula(1, 0)
formula_f2_deriv_x1 = f.getGradient().getImplementation().getFormula(0, 1)
formula_f2_deriv_x2 = f.getGradient().getImplementation().getFormula(1, 1)

# %%
# Then we create a new :class:`~openturns.SymbolicFunction` from these analytical expressions:
gradient_AsFunction = ot.SymbolicFunction(
    ["x1", "x2"],
    [
        formula_f1_deriv_x1,
        formula_f1_deriv_x2,
        formula_f2_deriv_x1,
        formula_f2_deriv_x2,
    ],
)
print(gradient_AsFunction)

# %%
# We want to define the hessian function :math:`\Rset^2 \rightarrow \Rset^6` defined by:
#
# .. math::
#    :label: hessianFct
#
#    \vect{x} = (x_1, x_2) \rightarrow & \left(\dfrac{\partial^2 f_1 (\vect{x})}{\partial x_1^2},
#                    \dfrac{\partial^2 f_1 (\vect{x})}{ \partial x_1\partial x_2},
#                    \dfrac{\partial^2 f_1 (\vect{x})}{\partial x_2^2},
#                    \dfrac{\partial f_2 (\vect{x})}{\partial x_1^2},
#                    \dfrac{\partial f_2 (\vect{x})}{ \partial x_1\partial x_2},
#                    \dfrac{\partial^2 f_2 (\vect{x})}{\partial x_2^2} \right)\\
#                & = (2, 0 , 4, 0, 0, 36x_2^2)
#
# To get the hessian as a function, we first get the analytical expressions of the second derivatives:
formula_f1_hessian_x1x1 = f.getHessian().getImplementation().getFormula(0, 0, 0)
formula_f1_hessian_x1x2 = f.getHessian().getImplementation().getFormula(1, 0, 0)
formula_f1_hessian_x2x2 = f.getHessian().getImplementation().getFormula(1, 1, 0)
formula_f2_hessian_x1x1 = f.getHessian().getImplementation().getFormula(0, 0, 1)
formula_f2_hessian_x1x2 = f.getHessian().getImplementation().getFormula(1, 0, 1)
formula_f2_hessian_x2x2 = f.getHessian().getImplementation().getFormula(1, 1, 1)

# %%
# Then we create a new :class:`~openturns.SymbolicFunction` from these analytical expressions:
hessian_AsFunction = ot.SymbolicFunction(
    ["x1", "x2"],
    [
        formula_f1_hessian_x1x1,
        formula_f1_hessian_x1x2,
        formula_f1_hessian_x2x2,
        formula_f2_hessian_x1x1,
        formula_f2_hessian_x1x2,
        formula_f2_hessian_x2x2,
    ],
)
print(hessian_AsFunction)

# %%
# Case 2: The general case
# ========================
#
# Here, we consider a function :math:`f` for which the analytical expressions
# of its gradient and Hessian are not available.
#
# To get the gradient function as a function defined in :eq:`gradientFct`, we have to use a
# :class:`~openturns.PythonFunction`. We re-use the previous function for for educational purposes.


def gradient_AsFunction_Python(inPoint):
    f1_deriv_x1 = f.gradient(inPoint)[0, 0]
    f1_deriv_x2 = f.gradient(inPoint)[1, 0]
    f2_deriv_x1 = f.gradient(inPoint)[0, 1]
    f2_deriv_x2 = f.gradient(inPoint)[1, 1]
    return [f1_deriv_x1, f1_deriv_x2, f2_deriv_x1, f2_deriv_x2]


gradient_AsFunction_OT = ot.PythonFunction(2, 4, gradient_AsFunction_Python)
print(gradient_AsFunction_OT([1.0, 2.0]))

# %%
# To get the hessian function as a function defined in :eq:`hessianFct`, we do the same:


def hessian_AsFunction_Python(inPoint):
    f1_hessian_x1x1 = f.hessian(inPoint)[0, 0, 0]
    f1_hessian_x1x2 = f.hessian(inPoint)[1, 0, 0]
    f1_hessian_x2x2 = f.hessian(inPoint)[1, 1, 0]
    f2_hessian_x1x1 = f.hessian(inPoint)[0, 0, 1]
    f2_hessian_x1x2 = f.hessian(inPoint)[1, 0, 1]
    f2_hessian_x2x2 = f.hessian(inPoint)[1, 1, 1]
    return [
        f1_hessian_x1x1,
        f1_hessian_x1x2,
        f1_hessian_x2x2,
        f2_hessian_x1x1,
        f2_hessian_x1x2,
        f2_hessian_x2x2,
    ]


hessian_AsFunction_OT = ot.PythonFunction(2, 6, hessian_AsFunction_Python)
print(hessian_AsFunction_OT([1.0, 2.0]))



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import sys

ot.TESTPREAMBLE()


class FUNC(ot.OpenTURNSPythonFunction):
    def __init__(self):
        super(FUNC, self).__init__(2, 1)
        self.setInputDescription(["R", "S"])
        self.setOutputDescription(["T"])

    def _exec(self, X):
        Y = [X[0] + X[1]]
        return Y


F = FUNC()
print(
    ("in_dim=" + str(F.getInputDimension()) + " out_dim=" + str(F.getOutputDimension()))
)
y = F((10, 5))
assert y == [15.0]
y = F(((10, 5), (6, 7)))
assert y == [[15.0], [13.0]]

# Instance creation
myFunc = ot.Function(F)

# Copy constructor
newFunc = ot.Function(myFunc)

assert myFunc.getInputDimension() == 2
assert myFunc.getOutputDimension() == 1

inPt = ot.Point(2, 2.0)
outPt = myFunc(inPt)
ott.assert_almost_equal(outPt, [4.0])

outPt = myFunc((10.0, 11.0))
ott.assert_almost_equal(outPt, [21.0])

inSample = ot.Sample(10, 2)
for i in range(10):
    inSample[i] = ot.Point((i, i))

outSample = myFunc(inSample)
print((repr(outSample)))
ott.assert_almost_equal(outSample, [[2 * i] for i in range(10)])

outSample = myFunc(((100.0, 100.0), (101.0, 101.0), (102.0, 102.0)))
print((repr(outSample)))
ott.assert_almost_equal(outSample, [[200], [202], [204]])

# test PythonFunction


def a_exec(X):
    Y = [0]
    Y[0] = X[0] + X[1]
    return Y


def a_exec_sample(Xs):
    Ys = []
    for X in Xs:
        Ys.append([X[0] + X[1]])
    return Ys


a_sample = ((100.0, 100.0), (101.0, 101.0), (102.0, 102.0))

print("exec")
myFunc = ot.PythonFunction(2, 1, a_exec)
outSample = myFunc(a_sample)
print(outSample)
ott.assert_almost_equal(outSample, [[200], [202], [204]])

print("exec + exec_sample")
myFunc = ot.PythonFunction(2, 1, a_exec, a_exec_sample)
outSample = myFunc(a_sample)
print(outSample)
ott.assert_almost_equal(outSample, [[200], [202], [204]])

print("exec_sample only on a point")
myFunc = ot.PythonFunction(2, 1, func_sample=a_exec_sample)
outSample = myFunc([100.0, 100.0])
print(outSample)
ott.assert_almost_equal(outSample, [200])

print("exec_sample only on a sample")
myFunc = ot.PythonFunction(2, 1, func_sample=a_exec_sample)
outSample = myFunc(a_sample)
print(outSample)
ott.assert_almost_equal(outSample, [[200], [202], [204]])

# multiprocessing spawn method on win (& osx for py>38) duplicates the output
cpus = -1 if sys.platform.startswith("linux") else None
print("distributed exec only on a point")
myFunc = ot.PythonFunction(2, 1, a_exec, n_cpus=cpus)
outSample = myFunc([100.0, 100.0])
print(outSample)
ott.assert_almost_equal(outSample, [200])

print("distributed exec only on a sample")
myFunc = ot.PythonFunction(2, 1, a_exec, n_cpus=cpus)
outSample = myFunc(a_sample)
print(outSample)
ott.assert_almost_equal(outSample, [[200], [202], [204]])


def a_grad(X):
    # wrong values but allows one to verify
    dY = [[1.0], [-1.0]]
    return dY


print("gradient")
myFunc = ot.PythonFunction(2, 1, a_exec, gradient=a_grad)
grad = myFunc.gradient([100.0, 100.0])
print(grad)


def wrong_dim_grad(X):
    # wrong dim: 3 rows, 1 column
    dY = [[1.0, -1.0, 42.0]]
    return dY


print("gradient with wrong dim")
myFunc = ot.PythonFunction(2, 1, a_exec, gradient=wrong_dim_grad)
with ott.assert_raises(RuntimeError):
    grad = myFunc.gradient([100.0, 100.0])


def a_hess(X):
    # wrong values but allows one to verify
    d2Y = [[[0.1], [0.3]], [[0.3], [0.1]]]
    return d2Y


print("hessian")
myFunc = ot.PythonFunction(2, 1, a_exec, hessian=a_hess)
hess = myFunc.hessian([100.0, 100.0])
print(hess)

print("no func")
try:
    myFunc = ot.PythonFunction(2, 1)
    outSample = myFunc(a_sample)
except Exception:
    # must raise exception
    print("no function detected : ok.")
else:
    raise Exception("no function not detected!")


def a_exec2(X):
    Y = [0]
    if X[0] == 0.0:
        Y[0] = bla  # noqa: F821
    elif X[0] == 1.0:
        Y[0] = "2" + 2
    elif X[0] == 2.0:
        Y[0] = 1 / 0
    return Y


myFunc = ot.PythonFunction(1, 1, a_exec2)
for x in [0.0, 1.0, 2.0]:
    try:
        X = [x]
        myFunc(X)
    except Exception as exc:
        # print exc
        print("exception handling: ok")
        if x == 0.0:
            assert "NameError" in str(exc)
        elif x == 1.0:
            assert "TypeError" in str(exc)
        elif x == 2.0:
            assert "ZeroDivisionError" in str(exc)


def a_exec3(X):
    X[8000]  # index error
    return [X[0] + X[1]]


myFunc = ot.PythonFunction(2, 1, a_exec3)
with ott.assert_raises(RuntimeError):
    myFunc([5, 6])

f = ot.PythonFunction(0, 3, lambda x: [42.0] * 3)
x = []
y = f(x)
print("y=", y)
ott.assert_almost_equal(y, [42.0] * 3)


class BFunction(ot.OpenTURNSPythonFunction):
    def __init__(self):
        # super().__init__(3, 1)
        pass

    def _exec(self, X):
        return [X[0] + X[1]]


bModel = BFunction()
with ott.assert_raises(RuntimeError):
    # when super is not called in init, should not crash
    bFunction = ot.Function(bModel)



#! /usr/bin/env python

import openturns as ot
import math as m

ot.TESTPREAMBLE()


# DEFAULT CONSTRUCTOR AND STRING CONVERTER
print("test 0 : default constructor and string converter")

# Default constructor
matrix0 = ot.ComplexMatrix()

# String converter
print("matrix0 = ", repr(matrix0))

# CONSTRUCTOR WITH SIZE, OPERATOR() AND STRING CONVERTER
print("test number one : constructor with size, operator() and string converter")

# Constructor with size
matrix1 = ot.ComplexMatrix(2, 2)

# Check operator() methods
matrix1[0, 0] = 1.0 + 1j
matrix1[1, 0] = 2.0 + 4j
matrix1[0, 1] = 3.0 - 1j
matrix1[1, 1] = 4.0

# String converter
print("matrix1 = ", repr(matrix1))

# COPY CONSTRUCTOR AND STRING CONVERTER
print("test 2 : copy constructor and string converter")

# Copy constructor
matrix2 = ot.ComplexMatrix(matrix1)

# String converter
print("matrix2 = ", repr(matrix2))

# GET DIMENSIONS METHODS
print("test 3 : dimension methods")

# Get dimension methods
print("matrix1's nbRows = ", matrix1.getNbRows())
print("matrix1's nbColumns = ", matrix1.getNbColumns())

# CONSTRUCTOR WITH COLLECTION
print("test 4 : constructor with collection method")

# Create the collection of values
elementsValues = ot.ComplexCollection()
elementsValues.add(1.0 - 1j)
elementsValues.add(2.0 - 1j)
elementsValues.add(3.0 - 1j)
elementsValues.add(4.0 + 1j)
elementsValues.add(5.0 + 1j)
elementsValues.add(6.0 + 1j)

# Check the content of the collection
print("elementsValues = ", repr(elementsValues))

# Check the constructor with collection
matrix0bis = ot.ComplexMatrix(2, 2, elementsValues)
print("matrix0bis = ", repr(matrix0bis))

# TRANSPOSITION METHOD AND CONJUGATE METHOD
print("test 5 : transposition / conjugate method")

# Check transpose method
matrix4 = matrix1.transpose()
matrix5 = matrix1.conjugate()
print("matrix1 transposed = ", repr(matrix4))
print("matrix1 conjugated = ", repr(matrix5))

# TRANSPOSITION AND CONJUGATE COUPLED METHOD
print("transposition and conjugate method")

# Check transpose method
matrix6 = matrix1.conjugateTranspose()
print("matrix1 conjugated and transposed = ", repr(matrix6))

# ADDITION METHOD
print("test 6 : addition method")

# Check addition method : we check the operator and the symmetry of the
# operator, thus testing the comparison operator
sum1 = matrix1 + matrix4
sum2 = matrix4 + matrix1
print("sum1 = ", repr(sum1))
print("sum2 = ", repr(sum2))
print("sum1 equals sum2 = ", sum1 == sum2)

# SUBTRACTION METHOD
print("test 7 : subtraction method")

# Check subtraction method
diff = matrix1 - matrix4
print("diff = ", repr(diff))

#  MATRIX MULTIPLICATION METHOD
print("test 8 : matrix multiplication method")

# Check multiplication method
prod = matrix1 * matrix4
print("prod = ", repr(prod))

# MULTIPLICATION WITH A NUMERICAL POINT METHOD
print("test 9 : multiplication with a numerical point method")

# Create the numerical point
pt = ot.Point()
pt.add(1.0)
pt.add(2.0)
print("pt = ", repr(pt))

# Check the product method
ptResult = matrix1 * pt
print("ptResult = ", repr(ptResult))

#  MULTIPLICATION AND DIVISION BY A NUMERICAL SCALAR METHODS
print("test 10 : multiplication and division by a numerical scalar methods")

# Check the multiplication method
s = 3.0 + 1j
scalprod1 = matrix1 * s
print("scalprod1 = ", repr(scalprod1))

# Check the division method
scaldiv1 = matrix1 / s
print("scaldiv1 = ", repr(scaldiv1))

#  ISEMPTY METHOD
print("test 10 : isEmpty method")

# Check method isEmpty
matrix7 = ot.ComplexMatrix()
matrix8 = ot.ComplexMatrix()

print("matrix1 is empty = ", matrix1.isEmpty())
print("matrix5 is empty = ", matrix7.isEmpty())
print("matrix6 is empty = ", matrix8.isEmpty())
print("matrix0 is empty = ", matrix0.isEmpty())

# MULTIPLICATION WITH A NUMERICAL POINT METHOD
print("test 11 : multiplication with a numerical point method")

# Create the numerical point
pt_test = ot.Point()
pt_test.add(1.0)
pt_test.add(2.0)
print("pt_test = ", repr(pt_test))

A = ot.ComplexMatrix(2, 2)
A[0, 0] = 0.5
A[1, 0] = -(m.sqrt(3.0) / 2)
A[0, 1] = m.sqrt(3.0) / 2
A[1, 1] = 0.5
B = A.transpose()
identity = B * A

# Check the product method
ptResult2 = identity * pt_test
print("A = ", repr(A))
print("B = ", repr(B))
print("identity = ", repr(identity))
print("ptResult2 = ", repr(ptResult2))



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


dim = 3
size = 10
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * dim)
bounds = distribution.getRange()
lhs = ot.LHSExperiment(distribution, size)
design = lhs.generate()
subdivisions = [size] * dim

fig = plt.figure(figsize=(8, 8))
otv.PlotDesign(design, bounds, subdivisions, figure=fig)



"""
===========================
Create univariate functions
===========================
This example presents different ways to create univariate functions which can be used to create
a functional basis. This is the type of functions involved, for example, in the
:class:`~openturns.FunctionalChaosAlgorithm` class.

The univariate functions considered in this
example can be, in some cases, orthogonal to a distribution, but this is not a necessary
condition to create univariate functions. For example, Legendre polynomials are orthogonal
with respect to the uniform distribution, but the monomials of the canonical polynomial basis
are not necessarily orthogonal.
"""

# %%
# Description
# ===========
#
# We want to build univariate functions :math:`f : \mathbb{R} \mapsto \mathbb{R}`.
#
# We can do that:
#
# - Case 1: using orthogonal polynomials,
# - Case 2: using univariate polynomials,
# - Case 3: using orthogonal functions.
#
#
# Case 1: Orthogonal polynomials
# ==============================
#
# In that case, the polynomials are orthogonal with respect to a measure.
#
# For example: we consider the Legendre polynomials family, orthogonal with respect to the uniform
# distribution on :math:`[0,1]`. We use the :class:`~openturns.LegendreFactory` class.
# Its method :meth:`~openturns.LegendreFactory.build` applied to :math:`k` returns the polynomial number
# :math:`k` of the family.
import openturns as ot
import openturns.viewer as otv

f1 = ot.LegendreFactory().build(1)
f2 = ot.LegendreFactory().build(2)
f3 = ot.LegendreFactory().build(3)

print(type(f1))

g = f1.draw(-1.0, 1.0, 256)
g.add(f2.draw(-1.0, 1.0, 256))
g.add(f3.draw(-1.0, 1.0, 256))
g.setLegends([r"$\phi_1(x)$", r"$\phi_2(x)$", r"$\phi_3(x)$"])
g.setLegendPosition("lower right")
g.setTitle("Legendre Polynomials")
view = otv.View(g)

# %%
# We get the measure associated to the polynomial family:
measure_Legendre = ot.LegendreFactory().getMeasure()
print("Measure orthogonal to Legendre polynomials = ", measure_Legendre)

# %%
# Case 2: Univariate polynomials
# ==============================
#
# Univariate polynomials are not necessarily orthogonal with respect to a measure.
#
# We can use:
#
# - the :class:`~openturns.UniVariatePolynomial` class,
# - the :class:`~openturns.MonomialFunctionFactory` class,
# - the :class:`~openturns.MonomialFunction` class.
#
# For example, we consider :
#
# .. math::
#
#     \begin{array}{lcl}
#       f(x) & = & 1+2x+3x^2+4x^3\\
#       g(x) & = & x^3
#     \end{array}

f = ot.UniVariatePolynomial([1.0, 2.0, 3.0, 4.0])
g1 = ot.MonomialFunctionFactory().build(3)
g2 = ot.MonomialFunction(3)
print("f = ", f)
print("g1 = ", g1)
print("g2 = ", g2)

# %%
# There is no associated measure: if it is uncommented, the following command will fail, as expected.

# print(ot.MonomialFunctionFactory().getMeasure())

# %%
# Case 3: Orthogonal functions
# ============================
#
# In that case, the functions are orthogonal with respect to a measure :math:`\mu`.
#
# We can use:
#
# - the :class:`~openturns.HaarWaveletFactory` class,
# - the :class:`~openturns.FourierSeriesFactory` class.
#
# The method :meth:`~openturns.HaarWaveletFactory.build` returns the function number :math:`k` of the family.
#
# For example, we consider a Haar Wawelet.
f1 = ot.HaarWaveletFactory().build(1)
f2 = ot.HaarWaveletFactory().build(2)
f3 = ot.HaarWaveletFactory().build(3)

g = f1.draw(0.0, 1.0, 256)
g.add(f2.draw(0.0, 1.0, 256))
g.add(f3.draw(0.0, 1.0, 256))
g.setLegends([r"$\phi_1(x)$", r"$\phi_21(x)$", r"$\phi_3(x)$"])
g.setLegendPosition("upper right")
g.setTitle("Haar Wavelets")
view = otv.View(g)

# %%
# We get the measure: for the Haar Wavelet family, the :math:`\mathcal{U}(0,1)` distribution.
measure_Haar = ot.HaarWaveletFactory().getMeasure()
print("Measure orthogonal to Haar wavelets = ", measure_Haar)

# %%
# For example, we consider a Fourier Series.
f1 = ot.FourierSeriesFactory().build(1)
f2 = ot.FourierSeriesFactory().build(2)
f3 = ot.FourierSeriesFactory().build(3)

g = f1.draw(-3.0, 3.0, 256)
g.add(f2.draw(-3.0, 3.0, 256))
g.add(f3.draw(-3.0, 3.0, 256))
g.setLegends([r"$\phi_1(x)$", r"$\phi_21(x)$", r"$\phi_3(x)$"])
g.setLegendPosition("upper right")
g.setTitle("Fourier Series")
view = otv.View(g)

# %%
# We get the measure: for the Fourier Series, the :math:`\mathcal{U}(-\pi, \pi)` distribution.
measure_Fourier = ot.FourierSeriesFactory().getMeasure()
print("Measure orthogonal to Fourier series = ", measure_Fourier)


# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Beta(0.2, 0.4, -1.0, 2.0)
size = 10000
sample = distribution.getSample(size)
factory = ot.BetaFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
distribution = ot.Beta(0.5, 0.8, -1.0, 2.0)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
distribution = ot.Beta(0.5, 1.8, -1.0, 2.0)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
distribution = ot.Beta(1.5, 2.8, -1.0, 2.0)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedBeta = factory.buildAsBeta(sample)
print("Beta          =", distribution)
print("Estimated Beta=", estimatedBeta)
estimatedBeta = factory.buildAsBeta()
print("Default Beta=", estimatedBeta)
estimatedBeta = factory.buildAsBeta(distribution.getParameter())
print("Beta from parameters=", estimatedBeta)



#! /usr/bin/env python

import openturns as ot

# We create a Function
myFunc = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"],
    ["(x1*x1+x2^3*x1)/(2*x3*x3+x4^4+1)", "cos(x2*x2+x4)/(x1*x1+1+x3^4)"],
)

# We create a distribution
dim = myFunc.getInputDimension()
meanPoint = [0.2] * dim
sigma = [0.0] * dim
for i in range(dim):
    sigma[i] = 0.1 * (i + 1)
R = ot.CorrelationMatrix(dim)
for i in range(1, dim):
    R[i, i - 1] = 0.25

distribution = ot.Normal(meanPoint, sigma, R)

# We create a distribution-based RandomVector
X = ot.RandomVector(distribution)

# We create a composite RandomVector Y from X and myFunction
Y = ot.CompositeRandomVector(myFunc, X)

# We create a quadraticCumul algorithm
algo = ot.TaylorExpansionMoments(Y)

# We test the attributes here
print("algo=", algo)

# We compute the several elements provided by the quadratic cumul algorithm
print("First order mean=", repr(algo.getMeanFirstOrder()))
print("Second order mean=", repr(algo.getMeanSecondOrder()))
print("Covariance=", repr(algo.getCovariance()))
print("Value at mean=", repr(algo.getValueAtMean()))
print("Gradient at mean=", repr(algo.getGradientAtMean()))
print("Hessian at mean=", repr(algo.getHessianAtMean()))

algo_1 = ot.TaylorExpansionMoments(Y.getMarginal(0))
print("Importance factors=", repr(algo_1.getImportanceFactors()))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


chebychev = ot.ChebychevFactory()
print("chebychev=", chebychev)
for i in range(10):
    print("chebychev(", i, ")=", chebychev.build(i))
roots = chebychev.getRoots(10)
print("chebychev(10) roots=", repr(roots))
nodes, weights = chebychev.getNodesAndWeights(10)
print("chebychev(10) nodes=", nodes, "and weights=", weights)



import openturns as ot
from math import pi
import openturns.viewer as otv

# Create the function
formula = ["sin(X1) + 7. * sin(X2)^2 + 0.1 * X3^4 * sin(X1)"]
input_names = ["X1", "X2", "X3"]
g = ot.SymbolicFunction(input_names, formula)
# Create the probabilistic model
distributionList = [ot.Uniform(-pi, pi)] * 3
distribution = ot.JointDistribution(distributionList)
# Create a training sample
N = 500
inputTrain = distribution.getSample(N)
outputTrain = g(inputTrain)
# Create the chaos
multivariateBasis = ot.OrthogonalProductPolynomialFactory(distributionList)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
projectionStrategy = ot.LeastSquaresStrategy(
    inputTrain, outputTrain, selectionAlgorithm
)
totalDegree = 8
enumfunc = multivariateBasis.getEnumerateFunction()
P = enumfunc.getStrataCumulatedCardinal(totalDegree)
adaptiveStrategy = ot.FixedStrategy(multivariateBasis, P)
chaosalgo = ot.FunctionalChaosAlgorithm(
    inputTrain, outputTrain, distribution, adaptiveStrategy, projectionStrategy
)
chaosalgo.run()
result = chaosalgo.getResult()
# Draw Sobol' indices
chaosSI = ot.FunctionalChaosSobolIndices(result)
dim_input = distribution.getDimension()
first_order = [chaosSI.getSobolIndex(i) for i in range(dim_input)]
total_order = [chaosSI.getSobolTotalIndex(i) for i in range(dim_input)]
input_names = g.getInputDescription()
otv.View(
    ot.SobolIndicesAlgorithm.DrawSobolIndices(input_names, first_order, total_order)
)



#! /usr/bin/env python

import openturns as ot

# First, build two functions from R^3->R
inVar = ["x1", "x2", "x3"]
functions = list()
functions.append(
    ot.SymbolicFunction(
        inVar, ["x1^3 * sin(x2 + 2.5 * x3) - (x1 + x2)^2 / (1.0 + x3^2)"]
    )
)
functions.append(
    ot.SymbolicFunction(inVar, ["exp(-x1 * x2 + x3) / cos(1.0 + x2 * x3 - x1)"])
)
# Second, build the weights
coefficients = ot.Sample(0, 3)
coefficients.add([1.5, 2.5, -0.5])
coefficients.add([-3.5, 0.5, -1.5])
# Third, build the function
myFunction = ot.DualLinearCombinationFunction(functions, coefficients)
inPoint = ot.Point([1.2, 2.3, 3.4])
print("myFunction=", myFunction)
print("Value at ", inPoint, "=", myFunction(inPoint))
print("Gradient at ", inPoint, "=", myFunction.gradient(inPoint))
ot.PlatformInfo.SetNumericalPrecision(5)
print("Hessian at ", inPoint, "=", myFunction.hessian(inPoint))
for i in range(myFunction.getOutputDimension()):
    print("Marginal ", i, "=", myFunction.getMarginal(i))
print("Marginal (0,1)=", myFunction.getMarginal([0, 1]))
print("Marginal (0,2)=", myFunction.getMarginal([0, 2]))
print("Marginal (1,2)=", myFunction.getMarginal([1, 2]))

print("myFunction (HTML)=")
print(myFunction._repr_html_())



"""
Compute the L2 error between two functions
==========================================
"""

# %%

# %%
# In this example we compute the L2 error between two functions.

# %%
import openturns as ot
from openturns.usecases import ishigami_function

# %%
# This is easy using the arithmetic operator `-` on two
# instances of the :class:`~openturns.Function` class.

# %%
im = ishigami_function.IshigamiModel()
partOfIshigamiFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3"], ["7 * (sin(x2))^2 + 0.1 * x3^4 * sin(x1)"]
)
sampleSize = 1000
experiment = ot.MonteCarloExperiment(im.distribution, sampleSize)
integration = ot.ExperimentIntegration(experiment)
functionError = integration.computeL2Norm(im.model - partOfIshigamiFunction)
print(functionError)



import openturns as ot

ot.TESTPREAMBLE()

# Create the KL result
numberOfVertices = 20
interval = ot.Interval(-1.0, 1.0)
mesh = ot.IntervalMesher([numberOfVertices - 1]).build(interval)
covariance = ot.SquaredExponential()
process = ot.GaussianProcess(covariance, mesh)
sampleSize = 100
processSample = process.getSample(sampleSize)
threshold = 1.0e-7
algo = ot.KarhunenLoeveSVDAlgorithm(processSample, threshold)
algo.run()
klresult = algo.getResult()
print("ev=", klresult.getEigenvalues())
graph1 = klresult.drawEigenvalues()
graph2 = klresult.drawEigenvalues()
graph2.setLogScale(ot.GraphImplementation.LOGY)
graph3 = klresult.drawCumulatedEigenvaluesRemainder()
graph4 = klresult.drawCumulatedEigenvaluesRemainder()
graph4.setLogScale(ot.GraphImplementation.LOGY)
if 0:
    from openturns.viewer import View

    View(graph1).save("eigen1.png")
    View(graph2).save("eigen2.png")
    View(graph3).save("eigen3.png")
    View(graph4).save("eigen4.png")



#!/usr/bin/env python
"""
An optimization problem with mixed discrete input variables "4d soo".

min f(x1,x2,disc_b,logic)
where
x1 in [-5,5], x2 in [-3,3], disc_b in {True,False}, logic in {True,False}

4D function with a mixed decision space (two numeric, one discrete,
and one logical parameter)

In the original problem, we have disc_b in {"a","b"}.

Reference
Ported from R to Python.
Package ‘smoof’, August 14, 2017
Single and Multi-Objective Optimization Test Functions
Jakob Bossek
"""

import openturns as ot
import openturns.testing as ott


def fourdsoo(x):
    x1, x2, disc_b, logic = x
    # minimum : f([-5,-3,1,1]) = -18
    if round(disc_b) == 0.0:
        result = x1**2 + x2**2 + 10 * round(logic)
    else:
        result = x1 + x2 - 10 * round(logic)
    return [result]


# Create the objective function
objectiveFun = ot.PythonFunction(4, 1, fourdsoo)

x = [-5, -3, 1, 1]

print(f"f(x)={objectiveFun(x)}")

# Define bounds
bounds = ot.Interval([-5.0, -3.0, 0.0, 0.0], [5.0, 3.0, 1, 1])

# Define variable types
varTypes = [
    ot.OptimizationProblemImplementation.CONTINUOUS,
    ot.OptimizationProblemImplementation.CONTINUOUS,
    ot.OptimizationProblemImplementation.BINARY,
    ot.OptimizationProblemImplementation.INTEGER,
]

# Define problem
problem = ot.OptimizationProblem(objectiveFun)
problem.setVariablesType(varTypes)
problem.setBounds(bounds)
problem.setMinimization(True)

x0 = [0.0] * 4
algo = ot.Bonmin(problem)
algo.setStartingPoint(x0)
algo.setMaximumCallsNumber(10000)
algo.setMaximumIterationNumber(1000)

for name in ot.Bonmin.GetAlgorithmNames():
    if name in ["B-OA", "B-QG", "B-Ecp"]:
        continue
    print(f"-- {name} algorithm...")
    algo.setAlgorithmName(name)
    algo.run()
    result = algo.getResult()
    x_star = result.getOptimalPoint()
    print(f"x*={x_star}")
    y_star = result.getOptimalValue()
    neval = result.getCallsNumber()
    print(f"f(x*)={y_star} neval={neval}")
    ott.assert_almost_equal(x_star, [-5, -3, 1, 1], 5e-4)



#! /usr/bin/env python

from openturns.testing import assert_almost_equal
from openturns.usecases import wingweight_function

"""
Test the WingWeightModel data class on a reference point.
"""
ww = wingweight_function.WingWeightModel()


assert_almost_equal(ww.model(ww.distribution.getMean()), [267.625], 1e-2)



#! /usr/bin/env python
"""
@author: Michaël Baudin (EDF), Antoine Dumas (Phiméca)


Checks the distribution of the sensitivity indices estimators.

We compare the asymptotic distribution with an empirical sample of Sobol'
indices estimates.

We want to check that the distribution computed by the library is OK.
When we estimate this distribution depending on the asymptotic estimator,
this distribution is Gaussian.
Otherwise, it is computed using kernel smoothing based on a bootstrap re-sample.
A first significant point is the mean of the distribution.
The second significant point is the variance of the distribution, which
is computed using the Delta-method when the asymptotic distribution is
used.

The reference distribution is computed based on a sample of Sobol' indices,
which is generated by repetition.
Then we compare the "Sample" distribution (from repetition) and the
"Computed" distribution from the library.
These two distributions should be close.

The asymptotic distribution can be plot by two different methods.

- We can perform a single estimation of Sobol' indices (using a given sample size)
and get the estimated distribution.
The problem with this distribution is that it can be shifted with respect to the
true value of the Sobol' index.
This can be because of the limited sample size or because of a bias in the
estimator.
This is done if mean_distribution==True.
In this case, the figures are in the directory:
"graphe_validation_mean_distribution/".

- We can repeat the estimations several times and create a new Gaussian distribution,
which mean is computed as the mean of the means of the Gaussians and
the standard deviation is equal to the mean of the standard deviations.
This allows one to slightly correct the shifting generated by the limited size of the
sample, but does not correct the bias of the estimator, if any.
This is what is done if mean_distribution==False.
In this case the figures are in the the "graphe_validation/" directory.

The sample distribution of the Sobol' indices is estimated using an histogram,
based on a Monte-Carlo sample.

The check is based on the comparison between the distribution computed by the
library using a single sample and the distribution estimated by repetition.
"""


import matplotlib as mpl

import openturns as ot
import numpy as np
from openturns.viewer import View
from gsobollib import gsobolSAExact, gsobolDistribution, gsobol
from ishigamilib import (
    ishigamiSAExact,
    ishigamiDistribution,
    ishigamiAB,
    ishigamiGSymbolic,
)
from numpy import zeros, array
import pylab as pl
import pathlib
from sobol_variance_estimators import (
    SaltelliSensitivityAlgorithm,
    MartinezSensitivityAlgorithm,
    JansenSensitivityAlgorithm,
    MauntzKucherenkoSensitivityAlgorithm,
)

mpl.use("agg")


class SensitivityConfidenceTest:
    """
    Class pour valider la variance des estimateurs de sobols par la méthode Janon.
    La distribution asymptotique est comparé à la distribution empirique obtenue
    par répétitions du calcul des estimateurs de Sobol.
    """

    def __init__(
        self,
        model,
        distribution,
        sobol_estimator,
        FOexact=None,
        TOexact=None,
        sampleSize=500,
        seed=154681,
        nrepetitions=100,
        alpha=0.95,
        savefig=False,
        plot_figure=False,
    ):

        # Le modèle test : g-sobol ou ishigami
        self.model = model
        # la distribution des paramètres du modèle test
        self.distribution = distribution
        self.dim = distribution.getDimension()
        # classe améliorée pour le calcul des indices (par ex.
        # SaltelliSensitivityAlgorithm)
        self.sobol_estimator = sobol_estimator
        # Taille du plan d'expérience de base pour estimer S et ST
        self.sampleSize = sampleSize
        # Nombre de répétition de l'expérience
        self.nrepetitions = nrepetitions
        # bootstrap confidence level
        self.alpha = alpha
        # flag pour sauvegarder les figures
        self.savefig = savefig
        # first order exact indices if known
        self.FOexact = FOexact
        # total order exact indices if known
        self.TOexact = TOexact
        # seed
        self.seed = seed

        (
            sampleFirst,
            sampleTotal,
            foInterval,
            toInterval,
            distFirstCol,
            distTotalCol,
        ) = self.compute_sample_indices()
        if plot_figure:
            self.compare_last_repetition(
                sampleFirst,
                sampleTotal,
                foInterval,
                toInterval,
                distFirstCol,
                distTotalCol,
            )
            self.plot_indices_histogram(
                sampleFirst, sampleTotal, distFirstCol, distTotalCol
            )
            self.plot_indices_histogram(
                sampleFirst, sampleTotal, distFirstCol, distTotalCol, True
            )

    def compute_sample_indices(self):

        # Estimations des indices du premier ordre
        sampleFirst = zeros((self.nrepetitions, self.dim))

        # Estimations des indices totaux
        sampleTotal = zeros((self.nrepetitions, self.dim))

        # loi asymptotique
        distFirstCol = [object] * self.nrepetitions
        distTotalCol = [object] * self.nrepetitions

        # set seed of the random generator
        ot.RandomGenerator.SetSeed(self.seed)
        for i in range(self.nrepetitions):
            sobolexperiment = ot.SobolIndicesExperiment(
                self.distribution, int(self.sampleSize), False
            )
            inputDesign = sobolexperiment.generate()
            outputDesign = self.model(inputDesign)
            self.sensitivity_algorithm = self.sobol_estimator(
                inputDesign, outputDesign, int(self.sampleSize)
            )
            # self.sensitivity_algorithm = self.sobol_estimator(self.distribution,
            #                             int(self.sampleSize), self.model)
            self.sensitivity_algorithm.setConfidenceLevel(self.alpha)
            fo = self.sensitivity_algorithm.getAggregatedFirstOrderIndices()
            to = self.sensitivity_algorithm.getAggregatedTotalOrderIndices()
            # Récupère les distributions asymptotiques
            distFirstCol[i] = (
                self.sensitivity_algorithm.getFirstOrderAsymptoticDistribution()
            )
            distTotalCol[i] = (
                self.sensitivity_algorithm.getTotalOrderAsymptoticDistribution()
            )
            for j in range(self.dim):
                sampleFirst[i, j] = fo[j]
            for j in range(self.dim):
                sampleTotal[i, j] = to[j]

        # Récupère l'intervalle de confiance bootstrap pour le dernier échantillon
        foInterval = self.sensitivity_algorithm.getFirstOrderIndicesInterval()
        toInterval = self.sensitivity_algorithm.getTotalOrderIndicesInterval()

        # compute empirical variance
        self.std_first_empirical = ot.Sample(sampleFirst).computeStandardDeviation()
        self.std_total_empirical = ot.Sample(sampleTotal).computeStandardDeviation()
        return (
            sampleFirst,
            sampleTotal,
            foInterval,
            toInterval,
            distFirstCol,
            distTotalCol,
        )

    def compare_last_repetition(
        self,
        sampleFirst,
        sampleTotal,
        foInterval,
        toInterval,
        distFirstCol,
        distTotalCol,
    ):

        # récupère les valeurs des min et max des intervalles
        foIntervalMin = foInterval.getLowerBound()
        foIntervalMax = foInterval.getUpperBound()
        toIntervalMin = toInterval.getLowerBound()
        toIntervalMax = toInterval.getUpperBound()
        # Compare les intervalles bootstrap pour le dernier échantillon
        # et les quantiles issus des répétitions
        for j in range(self.dim):
            # Calcule les quantiles empiriques
            sampleFirstPerDim = ot.Sample([[v] for v in sampleFirst[:, j]])
            foMinj = sampleFirstPerDim.computeQuantile((1 - self.alpha) / 2)[0]
            foMaxj = sampleFirstPerDim.computeQuantile(1 - (1 - self.alpha) / 2)[0]
            sampleTotalPerDim = ot.Sample([[v] for v in sampleTotal[:, j]])
            toMinj = sampleTotalPerDim.computeQuantile((1 - self.alpha) / 2)[0]
            toMaxj = sampleTotalPerDim.computeQuantile(1 - (1 - self.alpha) / 2)[0]
            foAsympMinj = np.mean(
                [
                    distFirstCol[i][j].computeQuantile((1 - self.alpha) / 2)[0]
                    for i in range(self.nrepetitions)
                ]
            )
            foAsympMaxj = np.mean(
                [
                    distFirstCol[i][j].computeQuantile(1 - (1 - self.alpha) / 2)[0]
                    for i in range(self.nrepetitions)
                ]
            )
            toAsympMinj = np.mean(
                [
                    distTotalCol[i][j].computeQuantile((1 - self.alpha) / 2)[0]
                    for i in range(self.nrepetitions)
                ]
            )
            toAsympMaxj = np.mean(
                [
                    distTotalCol[i][j].computeQuantile(1 - (1 - self.alpha) / 2)[0]
                    for i in range(self.nrepetitions)
                ]
            )
            meanAsympFO = np.mean(
                [
                    distFirstCol[i][j].getStandardDeviation()[0]
                    for i in range(self.nrepetitions)
                ]
            )
            meanAsympTO = np.mean(
                [
                    distTotalCol[i][j].getStandardDeviation()[0]
                    for i in range(self.nrepetitions)
                ]
            )

            print("X%d" % (j))
            print(
                "   First standard deviation, Sample=%.5f, Asymptotic=%.5f"
                % (sampleFirstPerDim.computeStandardDeviation()[0], meanAsympFO)
            )
            print(
                "   Total standard deviation, Sample=%.5f, Asymptotic=%.5f"
                % (sampleTotalPerDim.computeStandardDeviation()[0], meanAsympTO)
            )
            print(
                "   First, Bootstrap=[%.4f,%.4f], Sample=[%.4f,%.4f], "
                "Asymptotic=[%.4f,%.4f]"
                % (
                    foIntervalMin[j],
                    foIntervalMax[j],
                    foMinj,
                    foMaxj,
                    foAsympMinj,
                    foAsympMaxj,
                )
            )
            print(
                "   Total, Bootstrap=[%.4f,%.4f], Sample=[%.4f,%.4f], "
                "Asymptotic=[%.4f,%.4f]"
                % (
                    toIntervalMin[j],
                    toIntervalMax[j],
                    toMinj,
                    toMaxj,
                    toAsympMinj,
                    toAsympMaxj,
                )
            )
            print("")

    def plot_indices_histogram(
        self,
        sampleFirst,
        sampleTotal,
        distFirstCol,
        distTotalCol,
        mean_distribution=False,
    ):
        fig, ax = pl.subplots(2, self.dim, figsize=(4 * self.dim, 8))
        fig.suptitle(
            "%s - N=%d - Repetitions = %d"
            % (self.sobol_estimator.__name__, self.sampleSize, self.nrepetitions)
        )

        # Pour chaque estimateur, compare la répartition empirique et la loi exacte
        for j in range(self.dim):
            # Indice du premier ordre
            sampleJ = sampleFirst[:, j]
            ax[0, j].hist(
                sampleJ,
                histtype="step",
                density=True,
                label="empirique - std=%.2e" % np.std(sampleJ),
            )
            if mean_distribution:
                # moyenne et ecart-type = moyenne des moyenne et écart-types
                # des lois asymptotiques
                mu = np.mean(
                    [distFirstCol[i][j].getMean()[0] for i in range(self.nrepetitions)]
                )  # valeur moyenne
                sigma = np.mean(
                    [
                        distFirstCol[i][j].getStandardDeviation()[0]
                        for i in range(self.nrepetitions)
                    ]
                )  # écart-type
                label = "asymptotique - std(mean)=%.2e" % sigma
            else:
                mu = distFirstCol[-1][j].getMean()[0]
                sigma = distFirstCol[-1][j].getStandardDeviation()[0]
                label = "asymptotique - std=%.2e" % sigma
            loiFo = ot.Normal(mu, sigma)
            View(loiFo.drawPDF(), axes=[ax[0, j]], plot_kwargs={"label": label})
            if self.FOexact is not None:
                ax[0, j].vlines(
                    self.FOexact[j],
                    0,
                    ax[0, j].get_ylim()[1],
                    label="exact - %0.3f" % self.FOexact[j],
                )
            ax[0, j].set_xlabel("S%d" % (j))
            ax[0, j].set_ylabel("Density")
            ax[0, j].legend(loc="lower right")
            # Indice du total
            sampleJ = sampleTotal[:, j]
            ax[1, j].hist(
                sampleJ,
                histtype="step",
                density=True,
                label="empirique - std=%.2e" % np.std(sampleJ),
            )
            if mean_distribution:
                # moyenne et ecart-type = moyenne des moyenne et écart-types
                # des lois asymptotiques
                mu = np.mean(
                    [distTotalCol[i][j].getMean()[0] for i in range(self.nrepetitions)]
                )  # valeur moyenne
                sigma = np.mean(
                    [
                        distTotalCol[i][j].getStandardDeviation()[0]
                        for i in range(self.nrepetitions)
                    ]
                )  # écart-type
                label = "asymptotique- std(mean)=%.2e" % sigma
            else:
                mu = distTotalCol[-1][j].getMean()[0]
                sigma = distTotalCol[-1][j].getStandardDeviation()[0]
                label = "asymptotique- std=%.2e" % sigma
            loiFo = ot.Normal(mu, sigma)
            View(loiFo.drawPDF(), axes=[ax[1, j]], plot_kwargs={"label": label})
            if self.TOexact is not None:
                ax[1, j].vlines(
                    self.TOexact[j],
                    0,
                    ax[1, j].get_ylim()[1],
                    label="exact - %0.3f" % self.TOexact[j],
                )
            ax[1, j].set_xlabel("ST%d" % (j))
            ax[1, j].set_ylabel("Density")
            ax[1, j].legend(loc="lower right")
        # fig.show()
        if mean_distribution:
            directory = "graphe_validation_mean_distribution/"
        else:
            directory = "graphe_validation/"
        # create directory if not exist
        pathlib.Path(directory).mkdir(parents=True, exist_ok=True)
        if self.savefig:
            fig.savefig(
                directory
                + "%s-%s.png" % (self.model.getName(), self.sobol_estimator.__name__),
                transparent=True,
                bbox_inches="tight",
            )


if __name__ == "__main__":

    # GSOBOL

    a = array([0, 9, 99])
    # Distribution uniforme associée au cas-test GSobol
    distribution_gsobol = gsobolDistribution(len(a))
    model_gsobol = ot.PythonFunction(len(a), 1, func_sample=lambda X: gsobol(X, a))
    model_gsobol.setName("G-Sobol")
    # Indices de sensibilité exacts
    [muexact, vexact, sexact, stexact] = gsobolSAExact(a)

    sTest_gsobol_saltelli = SensitivityConfidenceTest(
        model_gsobol,
        distribution_gsobol,
        SaltelliSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    sTest_gsobol_jansen = SensitivityConfidenceTest(
        model_gsobol,
        distribution_gsobol,
        JansenSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    sTest_gsobol_mauntz = SensitivityConfidenceTest(
        model_gsobol,
        distribution_gsobol,
        MauntzKucherenkoSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    sTest_gsobol_martinez = SensitivityConfidenceTest(
        model_gsobol,
        distribution_gsobol,
        MartinezSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    # ISHIGAMI

    model_ishigami = ishigamiGSymbolic()
    model_ishigami.setName("Ishigami")
    distribution_ishigami = ishigamiDistribution()
    # Indices de sensibilité exacts
    a, b = ishigamiAB()
    meanY, varY, S1, S2, S3, ST1, ST2, ST3 = ishigamiSAExact(a, b)
    sexact = ot.Point([S1, S2, S3])
    stexact = ot.Point([ST1, ST2, ST3])

    sTest_ishigami_saltelli = SensitivityConfidenceTest(
        model_ishigami,
        distribution_ishigami,
        SaltelliSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    sTest_ishigami_jansen = SensitivityConfidenceTest(
        model_ishigami,
        distribution_ishigami,
        JansenSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    sTest_ishigami_mauntz = SensitivityConfidenceTest(
        model_ishigami,
        distribution_ishigami,
        MauntzKucherenkoSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    sTest_ishigami_martinez = SensitivityConfidenceTest(
        model_ishigami,
        distribution_ishigami,
        MartinezSensitivityAlgorithm,
        FOexact=sexact,
        TOexact=stexact,
        savefig=True,
        plot_figure=True,
    )

    # POUTRE
    model_poutre = ot.SymbolicFunction(
        ["L", "b", "h", "E", "F"], ["F * L^3 / (48 * E * b * h^3 / 12)"]
    )
    model_poutre.setName("poutre")
    L = ot.LogNormal()
    L.setParameter(ot.LogNormalMuSigmaOverMu()([5.0, 0.02, 0.0]))
    b = ot.LogNormal()
    b.setParameter(ot.LogNormalMuSigmaOverMu()([0.2, 0.05, 0.0]))
    h = ot.LogNormal()
    h.setParameter(ot.LogNormalMuSigmaOverMu()([0.4, 0.05, 0.0]))
    E = ot.LogNormal()
    E.setParameter(ot.LogNormalMuSigmaOverMu()([3e4, 0.12, 0.0]))
    F = ot.LogNormal()
    F.setParameter(ot.LogNormalMuSigmaOverMu()([0.1, 0.20, 0.0]))
    distribution_poutre = ot.JointDistribution([L, b, h, E, F])

    sTest_poutre_saltelli = SensitivityConfidenceTest(
        model_poutre,
        distribution_poutre,
        SaltelliSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    sTest_poutre_jansen = SensitivityConfidenceTest(
        model_poutre,
        distribution_poutre,
        JansenSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    sTest_poutre_mauntz = SensitivityConfidenceTest(
        model_poutre,
        distribution_poutre,
        MauntzKucherenkoSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    sTest_poutre_martinez = SensitivityConfidenceTest(
        model_poutre,
        distribution_poutre,
        MartinezSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    # Aggregated Sobol

    model_aggregated = ot.SymbolicFunction(
        ["X1", "X2", "X3"],
        ["2*X1 + X2 - 3*X3 + 0.3*X1*X2", "-5*X1 + 4*X2 - 0.8*X2*X3 + 2*X3"],
    )
    model_aggregated.setName("AggregatedSobol")
    distribution_aggregated = ot.JointDistribution([ot.Uniform()] * 3)

    sTest_poutre_saltelli = SensitivityConfidenceTest(
        model_aggregated,
        distribution_aggregated,
        SaltelliSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    sTest_poutre_jansen = SensitivityConfidenceTest(
        model_aggregated,
        distribution_aggregated,
        JansenSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    sTest_poutre_mauntz = SensitivityConfidenceTest(
        model_aggregated,
        distribution_aggregated,
        MauntzKucherenkoSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )

    sTest_poutre_martinez = SensitivityConfidenceTest(
        model_aggregated,
        distribution_aggregated,
        MartinezSensitivityAlgorithm,
        savefig=True,
        plot_figure=True,
    )



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
dimension = 3
meanPoint = ot.Point(dimension, 1.0)
meanPoint[0] = 0.5
meanPoint[1] = -0.5
sigma = ot.Point(dimension, 1.0)
sigma[0] = 2.0
sigma[1] = 3.0
R = ot.CorrelationMatrix(dimension)
for i in range(1, dimension):
    R[i, i - 1] = 0.5

# Create a collection of distribution
aCollection = ot.DistributionCollection()

aCollection.add(ot.Normal(meanPoint, sigma, R))
meanPoint += ot.Point(meanPoint.getDimension(), 1.0)
aCollection.add(ot.Normal(meanPoint, sigma, R))
meanPoint += ot.Point(meanPoint.getDimension(), 1.0)
aCollection.add(ot.Normal(meanPoint, sigma, R))

# Instantiate one distribution object
distribution = ot.Mixture(aCollection, ot.Point(aCollection.getSize(), 2.0))
print("Distribution ", repr(distribution))
print("Weights = ", repr(distribution.getWeights()))
weights = distribution.getWeights()
weights[0] = 2.0 * weights[0]
distribution.setWeights(weights)
print("After update, new weights = ", repr(distribution.getWeights()))
distribution = ot.Mixture(aCollection)
print("Distribution ", repr(distribution))

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", repr(oneRealization))

# Test for sampling
size = 1000
oneSample = distribution.getSample(size)
print("oneSample first=", repr(oneSample[0]), " last=", repr(oneSample[size - 1]))
print("mean=", repr(oneSample.computeMean()))
print("covariance=", repr(oneSample.computeCovariance()))

# Define a point
point = ot.Point(distribution.getDimension(), 1.0)
print("Point= ", repr(point))

# derivative of PDF with regards its arguments
DDF = distribution.computeDDF(point)
print("ddf     =", repr(DDF))

# PDF value
LPDF = distribution.computeLogPDF(point)
print("log pdf=%.6f" % LPDF)
PDF = distribution.computePDF(point)
print("pdf     =%.6f" % PDF)

x = 0.6
y = [0.2] * (dimension - 1)
print("conditional PDF%.6f" % distribution.computeConditionalPDF(x, y))
print("conditional CDF%.6f" % distribution.computeConditionalCDF(x, y))
print("conditional quantile%.6f" % distribution.computeConditionalQuantile(x, y))
pt = ot.Point([i + 1.5 for i in range(dimension)])
print(
    "sequential conditional PDF=", distribution.computeSequentialConditionalPDF(point)
)
resCDF = distribution.computeSequentialConditionalCDF(pt)
print("sequential conditional CDF(", pt, ")=", resCDF)
print(
    "sequential conditional quantile(",
    resCDF,
    ")=",
    distribution.computeSequentialConditionalQuantile(resCDF),
)
# derivative of the PDF with regards the parameters of the distribution
CDF = distribution.computeCDF(point)
print("cdf=%.6f" % CDF)
CCDF = distribution.computeComplementaryCDF(point)
print("ccdf=%.6f" % CCDF)
PDFgr = distribution.computePDFGradient(point)
assert distribution.computePDFGradient([point]).getDimension() == len(PDFgr)
print("pdf gradient=", PDFgr)

# derivative of the PDF with regards the parameters of the distribution
CDFgr = distribution.computeCDFGradient(point)
print("cdf gradient=", CDFgr)

# quantile
quantile = distribution.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)=%.6f" % distribution.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f"
    % distribution.computeSurvivalFunction(inverseSurvival)
)
# Confidence regions
if distribution.getDimension() <= 2:
    (
        interval,
        threshold,
    ) = distribution.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
    print("Minimum volume interval=", interval)
    print("threshold=", ot.Point(1, threshold))
    levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
    print("Minimum volume level set=", levelSet)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
    print("Bilateral confidence interval=", interval)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, False
    )
    print("Unilateral confidence interval (lower tail)=", interval)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, True
    )
    print("Unilateral confidence interval (upper tail)=", interval)
    print("beta=", ot.Point(1, beta))

mean = distribution.getMean()
print("mean=", repr(mean))
covariance = distribution.getCovariance()
print("covariance=", repr(covariance))
parameters = distribution.getParametersCollection()
print("parameters=", repr(parameters))
parameter = distribution.getParameter()
print("parameter=", parameter)
parameterDescription = distribution.getParameterDescription()
print("parameter description=", parameterDescription)
distribution.setParameter(parameter)
print("Standard representative=", distribution.getStandardRepresentative())

# Constructor with separate weights. Also check small weights removal
weights = [1.0e-20, 2.5, 32.0]
atoms = ot.DistributionCollection(
    [ot.Normal(1.0, 1.0), ot.Normal(2.0, 2.0), ot.Normal(3.0, 3.0)]
)
newMixture = ot.Mixture(atoms, weights)
print("newMixture pdf= %.12g" % newMixture.computePDF(2.5))
print("atoms kept in mixture=", newMixture.getDistributionCollection())
print("newMixture=", newMixture)

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.skipMoments()  # slow
validation.skipCorrelation()  # slow
validation.run()

# check for non stricly increasing CDF: must return the inf of [1, 2]
distribution = ot.Mixture([ot.Uniform(0.0, 1.0), ot.Uniform(2.0, 3.0)])
q = distribution.computeQuantile(0.5)[0]
print(f"q={q:.6f}")



import openturns as ot
import openturns.viewer as otv
import math as m


def flooding(X):
    L, B = 5.0e3, 300.0
    Q, K_s, Z_v, Z_m = X
    alpha = (Z_m - Z_v) / L
    H = (Q / (K_s * B * m.sqrt(alpha))) ** (3.0 / 5.0)
    return [H]


g = ot.PythonFunction(4, 1, flooding)
Q = ot.TruncatedDistribution(ot.Gumbel(558.0, 1013.0), ot.TruncatedDistribution.LOWER)
K_s = ot.Dirac(30.0)
Z_v = ot.Dirac(50.0)
Z_m = ot.Dirac(55.0)
inputRandomVector = ot.JointDistribution([Q, K_s, Z_v, Z_m])
nbobs = 100
inputSample = inputRandomVector.getSample(nbobs)
outputH = g(inputSample)
Hobs = outputH + ot.Normal(0.0, 0.1).getSample(nbobs)
Qobs = inputSample[:, 0]
thetaPrior = [20, 49, 51]
model = ot.ParametricFunction(g, [1, 2, 3], thetaPrior)
errorCovariance = ot.CovarianceMatrix([[0.5**2]])
sigma = ot.CovarianceMatrix(3)
sigma[0, 0] = 5.0**2
sigma[1, 1] = 1.0**2
sigma[2, 2] = 1.0**2
algo = ot.GaussianNonLinearCalibration(
    model, Qobs, Hobs, thetaPrior, sigma, errorCovariance
)
algo.run()
result = algo.getResult()
graph = result.drawObservationsVsPredictions()

otv.View(graph)



"""
Draw the empirical CDF
======================
"""

# %%
# In this example we are going to draw the empirical CDF of an unidimensional sample.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Then create a sample from a Gaussian distribution.

# %%
size = 100
normal = ot.Normal(1)
sample = normal.getSample(size)

# %%
# We draw the empirical CDF based on the :class:`~openturns.FiniteDiscreteDistribution` distribution.
# By default, the `drawCDF` method requires no input argument.

# %%
distribution = ot.FiniteDiscreteDistribution(sample)
graph = distribution.drawCDF()
view = otv.View(graph)

# %%
# If required, we can specify the interval that we want to draw.
# In the following example, these bounds are computed from the minimum and the maximum of the sample.

# %%
xmin = sample.getMin()[0] - 2.0
xmax = sample.getMax()[0] + 2.0
graph = ot.FiniteDiscreteDistribution(sample).drawCDF(xmin, xmax)
view = otv.View(graph)

# %%
otv.View.ShowAll()



"""
Estimate moments iteratively
============================
"""

# %%
# In this example, we use the :class:`~openturns.IterativeMoments` class
# to compute iterative statistics.
# This class stores central moments up to a prescribed order iteratively.
# Then several statistics based on the moments are available depending on the
# chosen order.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We first create a one-dimensional Gaussian random variable to generate data.
dim = 1
distNormal = ot.Normal(dim)

# %%
# Then we use the central moments up to order 4 with the
# :class:`~openturns.IterativeMoments` class by giving the order (here 4)
# and the dimension (here 1):

# %%
order = 4
iterMoments = ot.IterativeMoments(order, dim)


# %%
# We can now perform the simulations.
# The :class:`~openturns.IterativeMoments` object stores the central
# moments iteratively.
# We first increment the object with one :class:`~openturns.Point` at a time.
# At any given step the current mean is obtained thanks to
# the :meth:`~openturns.IterativeMoments.getMean` method and the
# current number of iterations is given by
# the :meth:`~openturns.IterativeMoments.getIterationNumber` method.
size = 2000
meanEvolution = ot.Sample()
for i in range(size):
    point = distNormal.getRealization()
    iterMoments.increment(point)
    meanEvolution.add(iterMoments.getMean())

# %%
# We display the evolution of the mean.
iterationSample = ot.Sample.BuildFromPoint(range(1, size + 1))
curve = ot.Curve(iterationSample, meanEvolution)
graph = ot.Graph("Evolution of the mean", "iteration nb", "mean")
graph.add(curve)
graph.setLogScale(ot.GraphImplementation.LOGX)
view = otv.View(graph)

# %%
# We can also increment with a :class:`~openturns.Sample`.

# %%
sample = distNormal.getSample(size)
iterMoments.increment(sample)

# %%
# We print the total number of iterations and the mean.
print("Total number of iteration: ", iterMoments.getIterationNumber())
print("Mean: ", iterMoments.getMean())

# %%
# For the order of the `iterMoments` object is 4, we also have access to
# other statistics such as the variance (order 2), the skewness (order 3)
# or the kurtosis (order 4). For instance, a specified order of 3 would leave
# only the variance and the skewness available.
print("Variance: ", iterMoments.getVariance())
print("Skewness: ", iterMoments.getSkewness())
print("Kurtosis: ", iterMoments.getKurtosis())

otv.View.ShowAll()



"""
Create a full or sparse polynomial chaos expansion
==================================================
"""

# %%
# In this example we create a global approximation of a model using
# polynomial chaos expansion based on a design of experiments.
# The goal of this example is to show how we can create a full or
# sparse polynomial chaos expansion depending on our needs
# and depending on the number of observations we have.
# In general, we should have more observations than parameters to estimate.
# This is why a sparse polynomial chaos may be interesting:
# by carefully selecting the coefficients we estimate,
# we may reduce overfitting and increase the predictions of the
# metamodel.

# %%
import openturns as ot


# %%
# Define the model
# ~~~~~~~~~~~~~~~~

# %%
# Create the function.
myModel = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"], ["1 + x1 * x2 + 2 * x3^2 + x4^4"]
)

# %%
# Create a multivariate distribution.
distribution = ot.JointDistribution(
    [ot.Normal(), ot.Uniform(), ot.Gamma(2.75, 1.0), ot.Beta(2.5, 1.0, -1.0, 2.0)]
)

# %%
# In order to create the PCE, we can specify the distribution of the
# input parameters.
# If not known, statistical inference can be used to select a possible
# candidate, and fitting tests can validate such an hypothesis.
# Please read :doc:`Fit a distribution from an input sample </auto_surrogate_modeling/polynomial_chaos/plot_chaos_build_distribution>`
# for an example of this method.

# %%
# Create a training sample
# ~~~~~~~~~~~~~~~~~~~~~~~~

# %%
# Create a pair of input and output samples.
sampleSize = 250
inputSample = distribution.getSample(sampleSize)
outputSample = myModel(inputSample)

# %%
# Build the orthogonal basis
# ~~~~~~~~~~~~~~~~~~~~~~~~~~

# %%
# In the next cell, we create the univariate orthogonal polynomial basis
# for each marginal.
inputDimension = inputSample.getDimension()
coll = [
    ot.StandardDistributionPolynomialFactory(distribution.getMarginal(i))
    for i in range(inputDimension)
]
enumerateFunction = ot.LinearEnumerateFunction(inputDimension)
productBasis = ot.OrthogonalProductPolynomialFactory(coll, enumerateFunction)

# %%
# We can achieve the same result using :class:`~openturns.OrthogonalProductPolynomialFactory`.
marginalDistributionCollection = [
    distribution.getMarginal(i) for i in range(inputDimension)
]
multivariateBasis = ot.OrthogonalProductPolynomialFactory(
    marginalDistributionCollection
)
multivariateBasis

# %%
# Create a full PCE
# ~~~~~~~~~~~~~~~~~

# %%
# Create the algorithm.
# We compute the basis size from the total degree.
# The next lines use the :class:`~openturns.LeastSquaresStrategy` class
# with default parameters (the default is the
# :class:`~openturns.PenalizedLeastSquaresAlgorithmFactory` class).
# This creates a full polynomial chaos expansion, i.e.
# we keep all the candidate coefficients produced by the enumeration
# rule.
# In order to create a sparse polynomial chaos expansion, we
# must use the :class:`~openturns.LeastSquaresMetaModelSelectionFactory`
# class instead.
#
totalDegree = 3
candidateBasisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
print("Candidate basis size = ", candidateBasisSize)
adaptiveStrategy = ot.FixedStrategy(productBasis, candidateBasisSize)
projectionStrategy = ot.LeastSquaresStrategy()
algo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, distribution, adaptiveStrategy, projectionStrategy
)
algo.run()
result = algo.getResult()
result

# %%
# Get the number of coefficients in the PCE.
selectedBasisSizeFull = result.getIndices().getSize()
print("Selected basis size = ", selectedBasisSizeFull)

# %%
# We see that the number of coefficients in the selected basis is
# equal to the number of coefficients in the candidate basis.
# This is, indeed, a *full* PCE.

# %%
# Use the PCE
# ~~~~~~~~~~~

# %%
# Get the metamodel function.
metamodel = result.getMetaModel()

# %%
# In order to evaluate the metamodel on a single point, we just
# use it as any other :class:`openturns.Function`.
xPoint = distribution.getMean()
yPoint = metamodel(xPoint)
print("Value at ", xPoint, " is ", yPoint)

# %%
# Based on these results, we may want to validate our metamodel.
# More details on this topic are presented in
# :doc:`Validate a polynomial chaos </auto_surrogate_modeling/polynomial_chaos/plot_chaos_draw_validation>`.

# %%
# Create a sparse PCE
# ~~~~~~~~~~~~~~~~~~~

# %%
# In order to create a sparse polynomial chaos expansion, we
# use the :class:`~openturns.LeastSquaresMetaModelSelectionFactory`
# class instead.
#
totalDegree = 6
candidateBasisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
print("Candidate basis size = ", candidateBasisSize)
adaptiveStrategy = ot.FixedStrategy(productBasis, candidateBasisSize)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
projectionStrategy = ot.LeastSquaresStrategy(selectionAlgorithm)
algo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, distribution, adaptiveStrategy, projectionStrategy
)
algo.run()
result = algo.getResult()
result

# %%
# Get the number of coefficients in the PCE.
selectedBasisSizeSparse = result.getIndices().getSize()
print("Selected basis size = ", selectedBasisSizeSparse)

# %%
# We see that the number of selected coefficients is lower than
# the number of candidate coefficients.
# This may reduce overfitting and can produce a PCE with more
# accurate predictions.



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import sys

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(3)


def progress(percent):
    sys.stderr.write("-- progress=" + str(percent) + "%\n")


def stop():
    sys.stderr.write("-- stop?\n")
    return False


dim = 2
f = ot.SymbolicFunction(["x1", "x2"], ["10-5*(x1-3)^2-7*(x2-2)^2+0.1*(x1+x2)"])
startingPoint = [0.0] * dim
bounds = ot.Interval([-6.0, -6.0], [1.0, 7.0])
algoNames = ot.NLopt.GetAlgorithmNames()

for algoName in algoNames:
    algo = ot.NLopt(algoName)
    if algoName == "GN_ISRES":
        algo.setMaximumConstraintError(1e-2)
    for minimization in [False, True]:
        for inequality in [False, True]:
            for equality in [False, True]:
                for bound in [False, True]:
                    if minimization and not bound:
                        continue
                    print(
                        "algo=",
                        algoName,
                        "minimization=",
                        minimization,
                        "bounds=",
                        bound,
                        "inequality=",
                        inequality,
                        "equality=",
                        equality,
                    )
                    problem = ot.OptimizationProblem(f)
                    problem.setMinimization(minimization)
                    if inequality:
                        # x1 <= 2
                        problem.setInequalityConstraint(
                            ot.SymbolicFunction(["x1", "x2"], ["2-x1"])
                        )
                    if equality:
                        # x2 = 4
                        problem.setEqualityConstraint(
                            ot.SymbolicFunction(["x1", "x2"], ["x2-4"])
                        )
                    if bound:
                        problem.setBounds(bounds)
                    try:
                        algo.setProblem(problem)
                    except Exception:
                        print("-- Not supported")
                        continue
                    algo.setMaximumCallsNumber(1000)
                    algo.setStartingPoint(startingPoint)
                    try:
                        algo.run()
                    except Exception as e:
                        print("-- ", e)
                        continue
                    result = algo.getResult()
                    x = result.getOptimalPoint()
                    print("x^=", x, "y^=", result.getOptimalValue())
                    if not inequality and not equality:
                        if not minimization:
                            if not bound:
                                x_ref = [3.0, 2.0]
                            else:
                                x_ref = [1.0, 2.0]
                        else:
                            x_ref = [-6.0, -6.0]
                            # these go to (-6,7)
                            if "NELDERMEAD" in algoName or "SBPLX" in algoName:
                                continue
                        ott.assert_almost_equal(x, x_ref, 4e-1, 1e-2)
                    elif equality:
                        ott.assert_almost_equal(x[1], 4.0, 4e-1, 1e-2)
                    elif inequality:
                        assert x[0] < 2.01, "!x1<=2.0"

# FORM
f = ot.SymbolicFunction(["E", "F", "L", "I"], ["-F*L^3/(3*E*I)"])
dim = f.getInputDimension()
mean = [50.0, 1.0, 10.0, 5.0]
sigma = ot.Point(dim, 1.0)
R = ot.IdentityMatrix(dim)
distribution = ot.Normal(mean, sigma, R)
vect = ot.RandomVector(distribution)
output = ot.CompositeRandomVector(f, vect)
event = ot.ThresholdEvent(output, ot.Less(), -3.0)
solver = ot.NLopt("LD_AUGLAG")
solver.setMaximumCallsNumber(400)
solver.setMaximumAbsoluteError(1.0e-10)
solver.setMaximumRelativeError(1.0e-10)
solver.setMaximumResidualError(1.0e-10)
solver.setMaximumConstraintError(1.0e-10)
solver.setStartingPoint(mean)
algo = ot.FORM(solver, event)
algo.run()
result = algo.getResult()
beta = result.getGeneralisedReliabilityIndex()
print("beta=%.6f" % beta)
ott.assert_almost_equal(beta, 1.009255)
